#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-09-27
# @Filename: status.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from clu import Command

from hal.macros.expose import ExposeMacro

from . import hal_command_parser


if TYPE_CHECKING:
    from . import HALCommandType


__all__ = ["status"]


@hal_command_parser.command()
@click.option("--full", is_flag=True, help="Outputs additional information.")
async def status(command: HALCommandType, full: bool = False):
    """Outputs the status of the system."""

    actor = command.actor

    await Command("script list", parent=command).parse()

    macros = actor.helpers.macros
    command.info(running_macros=[macro for macro in macros if macros[macro].running])

    if full:
        command.info(macros=sorted(list(macros)))
        for macro_name in sorted(macros):
            macros[macro_name].list_stages(command, level="d")
            macros[macro_name].output_stage_status(command, level="d")

    command.info(bypasses=list(command.actor.helpers.bypasses))

    expose_macro = macros["expose"]
    assert isinstance(expose_macro, ExposeMacro)
    command.debug(expose_is_paused=not expose_macro._pause_event.is_set())

    return command.finish(text="Alles ist gut.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-10-10
# @Filename: macro.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import enum
import warnings
from collections import defaultdict
from contextlib import suppress

from typing import TYPE_CHECKING, Any, ClassVar, Coroutine, Optional, Union

from clu import Command, CommandStatus

from hal import config
from hal.exceptions import HALUserWarning, MacroError


if TYPE_CHECKING:
    from hal.actor import HALCommandType


__all__ = ["Macro"]


StageType = Union[str, tuple[str, ...], list[str]]


class StageStatus(enum.Flag):
    """Stage status codes."""

    WAITING = enum.auto()
    ACTIVE = enum.auto()
    CANCELLED = enum.auto()
    CANCELLING = enum.auto()
    FINISHED = enum.auto()
    FAILED = enum.auto()


def flatten(stages: list[StageType]) -> list[str]:
    flat = []
    for stage in stages:
        if isinstance(stage, str):
            flat.append(stage)
        else:
            flat += list(stage)
    return flat


class Macro:
    """A base macro class that offers concurrency and cancellation."""

    __RUNNING__: ClassVar[list[str]] = []

    name: str

    __STAGES__: list[StageType]
    __PRECONDITIONS__: list[StageType] = []
    __CLEANUP__: list[StageType] = []

    def __init__(self, name: Optional[str] = None):
        if name is None and not hasattr(self, "name"):
            raise MacroError("The macro does not have a name attribute.")
        self.name = name or self.name

        if not hasattr(self, "__STAGES__"):
            raise MacroError("Must override __STAGES__.")

        self.stages = self.__PRECONDITIONS__ + self.__STAGES__ + self.__CLEANUP__

        self.flat_stages = flatten(self.stages)

        for stage in self.__PRECONDITIONS__ + self.__CLEANUP__:
            if not isinstance(stage, str):
                raise MacroError(
                    "Preconditions and cleanup stages cannot run in parallel."
                )
            if stage in flatten(self.__STAGES__):
                raise MacroError(f"Stage {stage} cannot be a selectable stage.")

        if len(self.stages) != len(set(self.stages)):
            raise MacroError("Duplicate stages found.")

        self.stage_status = {st: StageStatus.WAITING for st in flatten(self.stages)}

        self._base_config = config["macros"].get(self.name, {}).copy()
        self.config: defaultdict[str, Any] = defaultdict(
            lambda: None, self._base_config.copy()
        )

        self.command: HALCommandType

        self._running: bool = False
        self.failed: bool = False
        self.cancelled: bool = False

        self._running_task: asyncio.Task | None = None
        self._running_event = asyncio.Event()
        self._running_event.set()  # Won't unset until the macro is actually running.

    def __repr__(self):
        stages = flatten(self.stages)
        return f"<{self.__class__.__name__} (name={self.name}, stages={stages})>"

    def _reset_internal(self, **opts):
        """Internal reset method that can be overridden by the subclasses."""

        pass

    def reset(
        self,
        command: HALCommandType,
        reset_stages: Optional[list[StageType]] = None,
        force: bool = False,
        **opts,
    ):
        """Resets stage status.

        ``reset_stages`` is a list of stages to be executed when calling `.run`.
        If ``reset_stages`` is a list of string and ``force=False``, the reset
        stages are rearranged according to the original ``__STAGES__`` order,
        including concurrent stages. If the list of ``reset_stages`` includes
        tuples representing stages that must be run concurrently, or
        ``force=True``, then the stages are run as input.

        ``opts`` parameters will be used to override the macro configuration
        options only until the next reset.

        """

        if command is None:
            raise MacroError("A new command must be passed to reset.")

        self.command = command

        if reset_stages is None:
            self.stages = self.__PRECONDITIONS__ + self.__STAGES__ + self.__CLEANUP__
        else:
            self.stages = []

            if force is False and all([isinstance(x, str) for x in reset_stages]):
                for stage in self.__STAGES__:
                    if isinstance(stage, str) and stage in reset_stages:
                        self.stages.append(stage)
                    elif isinstance(stage, (tuple, list)):
                        if all([x in reset_stages for x in stage]):
                            self.stages.append(stage)
                        else:
                            for x in stage:
                                if x in reset_stages:
                                    self.stages.append(x)
            else:
                self.stages = reset_stages.copy()

            for stage in flatten(self.stages):
                if stage not in flatten(self.__STAGES__):
                    raise MacroError(f"Unknown stage {stage}.")

            for pre_stage in self.__PRECONDITIONS__:
                if pre_stage not in flatten(self.stages):
                    self.stages.insert(0, pre_stage)

            for cleanup_stage in self.__CLEANUP__:
                if cleanup_stage not in flatten(self.stages):
                    self.stages.append(cleanup_stage)

        if len(self.stages) == 0:
            raise MacroError("No stages found.")

        self.flat_stages = flatten(self.stages)

        # Reload the config and update it with custom options for this run.
        self.config = defaultdict(lambda: None, self._base_config.copy())
        self.config.update(
            {k: v for k, v in opts.items() if k not in self.config or v is not None}
        )

        self.failed = False
        self.cancelled = False

        self.stage_status = {st: StageStatus.WAITING for st in flatten(self.stages)}

        for st in self.stage_status:
            if getattr(self, st, None) is None:
                raise MacroError(f"Cannot find method for stage {st!r}.")

        self._reset_internal(**opts)

        self.running = False
        self._running_task = None

        if not self._running_event.is_set():
            self._running_event.set()

        self.list_stages()

    @property
    def actor(self):
        """Returns the command actor."""

        return self.command.actor

    @property
    def helpers(self):
        """Returns the actor helpers."""

        return self.actor.helpers

    @property
    def running(self):
        """Is the macro running?"""
        return self._running

    @running.setter
    def running(self, is_running: bool):
        """Sets the macro status as running/not-running and informs the actor."""

        if self._running == is_running:
            return

        self._running = is_running

        if is_running is True and self.name not in Macro.__RUNNING__:
            Macro.__RUNNING__.append(self.name)
        elif is_running is False and self.name in Macro.__RUNNING__:
            Macro.__RUNNING__.remove(self.name)

        self.command.debug(running_macros=Macro.__RUNNING__)

        if is_running:
            if not self._running_event.is_set():
                self._running_event.set()
            self._running_event.clear()
        else:
            self._running_event.set()

    def set_stage_status(
        self,
        stages: StageType,
        status: StageStatus,
        output: bool = True,
    ):
        """Set the stage status and inform the actor."""

        if isinstance(stages, str):
            stages = (stages,)

        for stage in stages:
            if stage not in self.stage_status:
                warnings.warn(
                    f"Cannot find stage {stage} in list. "
                    "Maybe the macro was not reset correctly.",
                    HALUserWarning,
                )
                return

            self.stage_status[stage] = status

        if output:
            self.output_stage_status()

    def output_stage_status(
        self,
        command: Optional[HALCommandType] = None,
        level: str = "d",
    ):
        """Outputs the stage status to the actor."""

        out_command = command or self.command

        status_keyw = [self.name]
        for stage in self.stage_status:
            status_name = self.stage_status[stage].name
            assert status_name
            status_keyw += [stage, status_name.lower()]

        out_command.write(level, stage_status=status_keyw)

    def list_stages(
        self,
        command: Optional[HALCommandType] = None,
        level: str = "i",
        only_all: bool = False,
    ):
        """Outputs stages to the actor."""

        list_command = command or self.command

        if only_all is False:
            list_command.write(
                level,
                stages=[self.name] + [st for st in flatten(self.stages)],
            )

        list_command.write(
            level,
            all_stages=[self.name]
            + [st for st in flatten(self.__STAGES__ + self.__CLEANUP__)],
        )

    async def fail_macro(
        self,
        error_or_message: Exception | str,
        stage: Optional[StageType] = None,
    ):
        """Fails the macros and informs the actor."""

        if isinstance(error_or_message, Exception):
            self.command.error(error=error_or_message)
            self.command.actor.log.exception(f"Macro {self.name} failed with error:")
        else:
            self.command.info(text=error_or_message)

        self.failed = True

        if stage is None:
            stage = list(self.stage_status.keys())
        else:
            if not isinstance(stage, (list, tuple)):
                stage = [stage]

        for ss in stage:
            if ss in self.__CLEANUP__:
                continue
            if self.stage_status[ss] == StageStatus.ACTIVE:
                self.stage_status[ss] = StageStatus.FAILED
            if self.stage_status[ss] == StageStatus.WAITING:
                self.stage_status[ss] = StageStatus.CANCELLED

        self.output_stage_status()

        for icleanup, cleanup_stage in enumerate(self.__CLEANUP__):
            if self.stage_status[str(cleanup_stage)] != StageStatus.WAITING:
                continue

            if icleanup == 0:
                self.command.warning("Running cleanup tasks.")

            try:
                self.set_stage_status(cleanup_stage, StageStatus.ACTIVE)
                await asyncio.gather(*self._get_coros(cleanup_stage))
                self.set_stage_status(cleanup_stage, StageStatus.FINISHED)
            except Exception as err:
                self.command.error(f"Cleanup {cleanup_stage} failed: {err}")
                self.set_stage_status(cleanup_stage, StageStatus.FAILED)
                break

        self.running = False

    async def run(self):
        """Executes the macro allowing for cancellation."""

        if self.running:
            raise MacroError("This macro is already running.")

        if not hasattr(self, "command") or self.command is None:
            raise MacroError("Cannot start a macro without an active command.")

        if self.command.status.is_done:
            raise MacroError("The command is done.")

        self.running = True
        self._running_task = asyncio.create_task(self._do_run())

        try:
            await self._running_task
        except asyncio.CancelledError:
            with suppress(asyncio.CancelledError):
                self._running_task.cancel()
                await self._running_task
        except Exception as err:
            await self.fail_macro(err)

        self.running = False

        return StageStatus.FAILED not in self.stage_status.values()

    def _get_coros(self, stage: StageType) -> list[Coroutine]:
        if isinstance(stage, str):
            stage_method = getattr(self, stage)

            if not asyncio.iscoroutinefunction(stage_method):
                raise MacroError(f"Stage function for {stage} is not a coroutine.")

            return [stage_method()]

        else:
            coros: list[Coroutine] = []
            for st in stage:
                coros += self._get_coros(st)

            return coros

    async def _do_run(self):
        """Actually run the stages."""

        current_task: asyncio.Future | None = None

        for istage, stage in enumerate(self.stages):
            stage_coros = [asyncio.create_task(coro) for coro in self._get_coros(stage)]
            current_task = asyncio.gather(*stage_coros)

            self.set_stage_status(stage, StageStatus.ACTIVE)

            try:
                await current_task

                was_cancelling = self.has_status(stage, StageStatus.CANCELLING)

                # Regardless of whether it was cancelling, this stage finished.
                self.set_stage_status(stage, StageStatus.FINISHED)

                # But now abort the macro.
                if was_cancelling:
                    raise asyncio.CancelledError()

            except asyncio.CancelledError:
                with suppress(asyncio.CancelledError):
                    current_task.cancel()
                    await current_task

                # Cancel this and all future stages.
                cancel_stages = [
                    stg
                    for stg in flatten(self.stages[istage:])
                    if stg not in self.__CLEANUP__
                    and self.stage_status[stg] != StageStatus.FINISHED
                ]
                self.set_stage_status(
                    cancel_stages,
                    StageStatus.CANCELLED,
                    output=False,
                )
                self.cancelled = True

                # Not really failing the macro since this was a user
                # requested cancellation.
                await self.fail_macro("The macro was cancelled")
                return

            except Exception as err:
                warnings.warn(
                    f"Macro {self.name} failed with error '{err}'",
                    HALUserWarning,
                )

                # Cancel stage tasks (in case we are running multiple concurrently).
                with suppress(asyncio.CancelledError):
                    for task in stage_coros:
                        if not task.done():
                            task.cancel()

                await self.fail_macro(err, stage=stage)
                return

    def get_active_stages(self):
        """Returns a list of running stages."""

        return [
            stage
            for stage in self.stage_status
            if self.stage_status[stage] == StageStatus.ACTIVE
        ]

    def is_stage_done(self, stage: str):
        """Returns `True` if a stage is finished."""

        return self.has_status(stage, StageStatus.FINISHED)

    def has_status(self, stages: StageType, status: StageStatus):
        """Determines if any of the stages has that status."""

        if isinstance(stages, str):
            return self.stage_status[stages] == status

        return any([self.stage_status[stage] == status for stage in stages])

    def cancel(self, now: bool = True):
        """Cancels the execution ot the macro."""

        if not self.running or not self._running_task:
            raise MacroError("The macro is not running.")

        if now:
            self._running_task.cancel()
        else:
            active_stages = self.get_active_stages()
            self.set_stage_status(active_stages, StageStatus.CANCELLING)

    async def send_command(
        self,
        target: str,
        command_string: str,
        raise_on_fail: bool = True,
        **kwargs,
    ):
        """Sends a command to an actor. Raises `.MacroError` if it fails.

        Parameters
        ----------
        target
            Actor to command.
        command_string
            Command string to pass to the actor.
        raise_on_fail
            If the command failed, raise an error.
        kwargs
            Additional parameters to pass to ``send_command()``.

        """

        if not isinstance(self.command, Command):
            raise MacroError("A command is required to use send_command().")

        command = await self.command.send_command(target, command_string, **kwargs)

        if command.status.did_fail and raise_on_fail:
            if command.status == CommandStatus.TIMEDOUT:
                raise MacroError(f"Command {target} {command_string} timed out.")
            else:
                raise MacroError(f"Command {target} {command_string} failed.")

        return command

    async def wait_until_complete(self):
        """Asynchronously blocks until the macros is done, cancelled, or failed."""

        await self._running_event.wait()

        return not self.failed



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-16
# @Filename: ffs.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import enum

from typing import TYPE_CHECKING

from hal import config

from . import HALHelper


if TYPE_CHECKING:
    from hal.actor import HALCommandType


__all__ = ["FFSHelper"]


class FFSHelper(HALHelper):
    """Command and keeps track of the Flat-Field Screens status."""

    TIMEOUT: float = config["timeouts"]["ffs"]

    name = "ffs"

    def get_values(self):
        """Returns the FFS status flags."""

        values = self.actor.models["mcp"]["ffsStatus"].value
        if len(values) == 0 or all([value is None for value in values]):
            return [FFSStatus.UNKNWON] * 8

        return [FFSStatus(value) for value in values]

    def all_closed(self):
        """Returns `True` if all the petals are closed."""

        return all([x == FFSStatus.CLOSED for x in self.get_values()])

    def all_open(self):
        """Returns `True` if all the petals are open."""

        return all([x == FFSStatus.OPEN for x in self.get_values()])

    async def open(self, command: HALCommandType):
        """Open all the petals."""

        if self.all_open():
            return

        return await self._send_command(command, "mcp", "ffs.open")

    async def close(self, command: HALCommandType):
        """Close all the petals."""

        if self.all_closed():
            return

        return await self._send_command(command, "mcp", "ffs.close")


class FFSStatus(enum.Enum):
    """FFS status flags."""

    UNKNWON = "00"
    CLOSED = "01"
    OPEN = "10"
    INVALID = "11"



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-12-25
# @Filename: jaeger.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
from collections import deque
from dataclasses import dataclass

from typing import TYPE_CHECKING

from sdssdb.peewee.sdss5db import targetdb

from hal.helpers import HALHelper


if TYPE_CHECKING:
    from hal.actor import HALActor, HALCommandType


__all__ = ["JaegerHelper"]


@dataclass
class Configuration:
    """Stores information about a configuration."""

    actor: HALActor
    design_id: int
    configuration_id: int | None = None
    field_id: int | None = None
    cloned: bool = False
    preloaded: bool = False
    is_rm_field: bool = False
    new_field: bool = True
    observed: bool = False
    goto_complete: bool = False

    def __post_init__(self):
        if self.design_id is None or self.design_id < 0:
            return

        if self.check_db():
            self.set_field_id()

    def warn(self, message: str):
        """Warns users."""

        assert self.actor
        self.actor.write("w", error=message)

    def check_db(self):
        """Checks the database connection or tries to recreate it."""

        if targetdb.database.connected is False:
            self.warn("Cannot connect to database. Reconnecting.")

            if not targetdb.database.connect():
                self.warn(
                    "Cannot connect to database. "
                    "Field information may be incomplete.",
                )
                return False

        return True

    def set_field_id(self):
        """Finds the field_id for a design and whether it's RM/AQMES."""

        if self.field_id is None:
            try:
                self.field_id = targetdb.Design.get_by_id(self.design_id).field.field_id
            except Exception as err:
                self.warn(f"Failed getting field_id: {err}")
                return

        # Determine if a field is RM/AQMES. Worst case, just assume it is not.
        self.is_rm_field = False

        try:
            design_mode_label = (
                targetdb.Design.select(targetdb.Design.design_mode)
                .where(targetdb.Design.design_id == self.design_id)
                .scalar()
            )
            if design_mode_label is None:
                self.warn(f"Cannot find design_mode_label for design {self.design_id}")
            elif design_mode_label in ["dark_monit", "dark_rm"]:
                self.is_rm_field = True
        except Exception as err:
            self.warn(f"Failed determining RM/AQMES: {err}")

        return


class JaegerHelper(HALHelper):
    """Helper to interact with jaeger."""

    name = "jaeger"

    def __init__(self, actor: HALActor):
        super().__init__(actor)

        # A queue of previously loaded configurations. Really we
        # only care about the last one, but who knows?
        self._previous: deque[Configuration] = deque(maxlen=10)

        self.configuration: Configuration | None = None
        self.preloaded: Configuration | None = None

        self.model = self.actor.models["jaeger"]
        self.model["configuration_loaded"].register_callback(self._configuration_loaded)
        self.model["design_preloaded"].register_callback(self._design_preloaded)

    def warn(self, message: str):
        """Warns users."""

        self.actor.write("w", error=message)

    async def load_from_queue(self, command: HALCommandType, preload: bool = False):
        """(Pre-)Loads a design from the queue."""

        verb = "preload" if preload else "load"

        cmd = await self._send_command(
            command,
            "jaeger",
            f"configuration {verb} --epoch-delay 600",
            raise_on_fail=False,
        )

        if cmd.status.did_fail:
            self.warn("Failed loading design from the queue.")
            return False

        return True

    async def from_preloaded(self, command: HALCommandType):
        """Loads a preloaded design."""

        if self.preloaded is None:
            self.warn(
                "No active preloaded design. Will try to load a "
                "preloaded design but it will likely fail.",
            )

        await self._send_command(
            command,
            "jaeger",
            "configuration load --from-preloaded",
        )

    async def _configuration_loaded(self, key):
        """Processes a new configuration loaded."""

        current = self.configuration

        configuration_id = key.value[0]
        design_id = key.value[1]
        field_id = key.value[2]
        is_cloned = key.value[9]

        # First check if we had already preloaded this design.
        if self.preloaded and self.preloaded.design_id == design_id:
            new = self.preloaded

            new.preloaded = False
            new.configuration_id = configuration_id
            new.cloned = is_cloned
        else:
            new = Configuration(
                self.actor,
                design_id,
                configuration_id=configuration_id,
                field_id=field_id,
                cloned=is_cloned,
                preloaded=False,
            )

        user_message = (
            f"Found configuration with design_id={design_id}, field_id={field_id}."
        )

        if current:
            if current.observed and current.field_id == new.field_id:
                new.new_field = False
                user_message += " This is a repeat field design."

            self._previous.append(current)

        self.configuration = new
        self.preloaded = None

        self.actor.write("d", text=user_message)

    async def _design_preloaded(self, key):
        """Processes a new preloaded design."""

        design_id = key[0]

        # design_preloaded=-999 is output when the design is actually loaded.
        if design_id < 0:
            self.preloaded = None
            return

        # Wait for a tiny bit since the preloaded_is_cloned key is emitted at the
        # same time as design_preloaded and we want to be sure it has updated.
        # Worst case, it's not critical because when the real configuration is loaded
        # that will be updated.
        await asyncio.sleep(0.5)

        cloned = self.model["preloaded_is_cloned"].value[0]

        self.preloaded = Configuration(
            self.actor,
            design_id,
            cloned=cloned,
            preloaded=True,
        )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-10-10
# @Filename: goto_field.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
from time import time

from hal import config
from hal.exceptions import HALError, MacroError
from hal.helpers.lamps import LampsHelper
from hal.macros import Macro


__all__ = ["GotoFieldMacro"]


class GotoFieldMacro(Macro):
    """Go to field macro."""

    name = "goto_field"

    __PRECONDITIONS__ = ["prepare"]
    __STAGES__ = [
        ("slew", "reconfigure"),
        "fvc",
        ("reslew", "lamps"),
        "boss_flat",
        "boss_hartmann",
        "boss_arcs",
        "acquire",
        "guide",
    ]
    __CLEANUP__ = ["cleanup"]

    _lamps_task: asyncio.Task | None = None

    async def prepare(self):
        """Check configuration and run pre-slew checks."""

        self._lamps_task = None

        stages = self.flat_stages

        if "reconfigure" in stages:
            configuration_loaded = self.actor.models["jaeger"]["configuration_loaded"]
            last_seen = configuration_loaded.last_seen

            if last_seen is None:
                self.command.warning("The age of the loaded configuration is unknown.")
            elif time() - last_seen > 3600:  # One hour
                raise MacroError("Configuration is too old. Load a new configuration.")

        do_fvc = "fvc" in stages
        do_flat = "boss_flat" in stages
        do_arcs = "boss_hartmann" in stages or "boss_arcs" in stages

        # Stop the guider.
        # TODO: this will probably be different at LCO.
        if do_fvc or do_flat or do_arcs:
            await self.helpers.cherno.stop_guiding(self.command)
            await self.helpers.tcc.axis_stop(self.command)

        # Ensure the APOGEE shutter is closed but don't wait for it.
        asyncio.create_task(
            self.helpers.apogee.shutter(
                self.command,
                open=False,
                shutter="apogee",
            )
        )

        # Start closing the FFS if they are open but do not block. Only close the FFS
        # if we're going to do BOSS cals, otherwise it's about 20 seconds of lost time.
        if do_flat or do_arcs:
            await self._close_ffs(wait=False)

        # If lamps are needed, turn them on now but do not wait for them to warm up.
        # Do not turn lamps if we are going to take an FVC image. We add a delay
        # since the APOGEE shutter is closing and we don't want to start turning on
        # the lamps until it's fully closed.
        if do_flat and not do_fvc:
            self._lamps_task = asyncio.create_task(
                self.helpers.lamps.turn_lamp(
                    self.command,
                    ["ff"],
                    True,
                    turn_off_others=True,
                    delay=10,
                )
            )
        elif not do_flat and do_arcs:
            if do_fvc:
                lamps = ["HgCd"]
            else:
                lamps = ["HgCd", "Ne"]

            self._lamps_task = asyncio.create_task(
                self.helpers.lamps.turn_lamp(
                    self.command,
                    lamps,
                    True,
                    turn_off_others=True,
                    delay=10,
                )
            )
        else:
            await self._all_lamps_off(wait=False)

    async def slew(self):
        """Slew to field but keep the rotator at a fixed position."""

        # Wait five seconds to give time for the axis stop we issued in Prepare to
        # take effect.
        await asyncio.sleep(5)

        configuration_loaded = self.actor.models["jaeger"]["configuration_loaded"]
        ra, dec, pa = configuration_loaded[3:6]

        if any([ra is None, dec is None, pa is None]):
            raise MacroError("Unknown RA/Dec/PA coordinates for field.")

        result = self.actor.helpers.tcc.axes_are_clear()
        if not result:
            raise HALError("Some axes are not clear. Cannot continue.")

        # The fixed position at which to slew for the FVC loop.
        alt = self.config["fvc_alt"]
        az = self.config["fvc_az"]
        rot = self.config["fvc_rot"]
        keep_offsets = self.config["keep_offsets"]

        if self.config["fixed_rot"] is False:
            self.command.info("Slewing to field RA/Dec/PA.")
            await self.actor.helpers.tcc.goto_position(
                self.command,
                {"ra": ra, "dec": dec, "rot": pa},
                rotwrap="middle",
                keep_offsets=keep_offsets,
            )
        else:
            if self.config["fixed_altaz"]:
                self.command.info("Slewing to field with fixed rotator angle.")
                track_command = f"track {az}, {alt} mount /rota={rot} /rottype=mount"

            else:
                self.command.info("Slewing to field with fixed alt/az/rot position.")
                track_command = f"track {ra}, {dec} icrs /rota={rot} /rottype=mount"

            slew_result = await self.actor.helpers.tcc.do_slew(
                self.command,
                track_command=track_command,
                keep_offsets=keep_offsets,
            )

            if slew_result is False:
                raise HALError("Failed slewing to position.")

            self.command.info(text="Position reached.")

        self.command.info("Halting the rotator.")
        await self.helpers.tcc.axis_stop(self.command, axis="rot")

    async def reconfigure(self):
        """Reconfigures the FPS."""

        self.command.info("Reconfiguring FPS array.")

        # This is always safe. If it's already folded jaeger will return immediately.
        await self.send_command("jaeger", "configuration reverse")

        await self.send_command("jaeger", "configuration execute")

    async def fvc(self):
        """Run the FVC loop."""

        if "slew" in self.flat_stages:
            self.command.info("Halting the rotator.")
            await self.helpers.tcc.axis_stop(self.command, axis="rot")

        self.command.info("Running FVC loop.")
        fvc_command = await self.send_command(
            "jaeger",
            "fvc loop",
            time_limit=config["timeouts"]["fvc"],
            raise_on_fail=False,
        )

        if fvc_command.status.did_fail:
            raise MacroError("FVC loop failed.")

    async def reslew(self):
        """Re-slew to field."""

        configuration_loaded = self.actor.models["jaeger"]["configuration_loaded"]
        ra, dec, pa = configuration_loaded[3:6]

        if any([ra is None, dec is None, pa is None]):
            raise MacroError("Unknown RA/Dec/PA coordinates for field.")

        self.command.info("Re-slewing to field.")

        # Start going to position asynchronously.
        await self.actor.helpers.tcc.goto_position(
            self.command,
            {"ra": ra, "dec": dec, "rot": pa},
            rotwrap="nearest",
            keep_offsets=self.config["keep_offsets"],
        )

    async def lamps(self):
        """Ensures the correct lamps for calibrations are on."""

        cal_stages = ["boss_flat", "boss_hartmann", "boss_arcs"]
        if all([stage not in self.flat_stages for stage in cal_stages]):
            return

        # If we are going to take BOSS cals, start warming up lamps now (some may
        # already be on).
        if "boss_flat" in self.flat_stages:
            mode = "flat"
        elif "boss_hartmann" in self.flat_stages:
            mode = "hartmann"
        else:
            mode = "arcs"

        await self._ensure_lamps(mode)

    async def boss_flat(self):
        """Takes the BOSS flat."""

        self.command.info("Taking BOSS flat.")

        await self._ensure_lamps(mode="flat")

        # Now take the flat. Do not read it yet.
        flat_time = self.config["flat_time"]

        self.command.debug("Starting BOSS flat exposure.")

        await self.helpers.boss.expose(
            self.command,
            flat_time,
            exp_type="flat",
            readout=True,
            read_async=True,
        )

        asyncio.create_task(self.helpers.lamps.turn_lamp(self.command, "ff", False))

    async def boss_hartmann(self):
        """Takes the hartmann sequence."""

        await self._ensure_lamps(mode="hartmann")

        if self.helpers.boss.readout_pending:  # Potential readout from the flat.
            self.command.info("Waiting for BOSS to read out.")
            await self.helpers.boss.readout(self.command)

        # Run hartmann and adjust the collimator but ignore residuals.
        self.command.info("Running hartmann collimate.")
        await self.send_command(
            "hartmann",
            "collimate ignoreResiduals",
            time_limit=config["timeouts"]["hartmann"],
        )

        # Now check if there are residuals that require modifying the blue ring.
        sp1Residuals = self.actor.models["hartmann"]["sp1Residuals"][2]
        if sp1Residuals != "OK":
            raise MacroError(
                "Please adjust the blue ring and run goto-field again. "
                "The collimator has been adjusted."
            )

        if "boss_arcs" not in self.flat_stages:
            await self._all_lamps_off(wait=False)

    async def boss_arcs(self):
        """Takes BOSS arcs."""

        await self._ensure_lamps(mode="arcs")

        if self.helpers.boss.readout_pending:
            self.command.info("Waiting for BOSS to read out.")
            await self.helpers.boss.readout(self.command)

        self.command.info("Taking BOSS arc.")

        arc_time = self.config["arc_time"]

        await self.helpers.boss.expose(
            self.command,
            arc_time,
            exp_type="arc",
            readout=True,
            read_async=True,
        )

        await self._all_lamps_off(wait=False)

    async def acquire(self):
        """Acquires the field."""

        if self.helpers.cherno.is_guiding():
            self.command.info("Already guiding.")
            return

        pretasks = []

        if "reslew" not in self.flat_stages:
            self.command.info("Re-slewing to field.")
            pretasks.append(self.reslew())

        if not self.helpers.ffs.all_open():
            self.command.info("Opening FFS")
            pretasks.append(self.helpers.ffs.open(self.command))

        # Open FFS and re-slew at the same time.
        await asyncio.gather(*pretasks)

        # A bit of delay to make sure the axis status keyword is updated.
        await asyncio.sleep(1)

        if not self.helpers.tcc.check_axes_status("Tracking"):
            raise MacroError("Axes must be tracking for acquisition.")

        guider_time = self.config["guider_time"]
        target_rms = self.config["target_rms"]
        max_iterations = self.config["max_iterations"]

        self.command.info("Acquiring field.")
        await self.helpers.cherno.acquire(
            self.command,
            exposure_time=guider_time,
            target_rms=target_rms,
            max_iterations=max_iterations,
        )

    async def guide(self):
        """Starts the guide loop."""

        if self.helpers.cherno.is_guiding():
            self.command.info("Already guiding.")
            return

        if "acquire" not in self.flat_stages:
            self.command.info("Re-slewing to field.")
            await self.reslew()

        if not self.helpers.tcc.check_axes_status("Tracking"):
            raise MacroError("Axes must be tracking for guiding.")

        if not self.helpers.ffs.all_open():
            self.command.info("Opening FFS")
            await self.helpers.ffs.open(self.command)

        guider_time = self.config["guider_time"]

        self.command.info("Starting guide loop.")
        await self.helpers.cherno.guide(
            self.command,
            exposure_time=guider_time,
            wait=False,
        )

    async def cleanup(self):
        """Turns off all lamps."""

        # If enough stages have run, mark this configuration as goto_complete.
        if self.helpers.jaeger.configuration is not None and self._is_goto_complete():
            self.helpers.jaeger.configuration.goto_complete = True

        if self._lamps_task is not None and not self._lamps_task.done():
            self._lamps_task.cancel()

        await self.helpers.lamps.all_off(self.command, force=True)

        # Read any pending BOSS exposure.
        if self.helpers.boss.readout_pending:
            await self.helpers.boss.readout(self.command)

    async def _close_ffs(self, wait: bool = True):
        """Closes the FFS."""

        if not self.helpers.ffs.all_closed():
            self.command.info("Closing FFS")
            task = self.helpers.ffs.close(self.command)

            if wait:
                await task
            else:
                asyncio.create_task(task)

    async def _all_lamps_off(self, wait: bool = True):
        """Turns all the lamps off after checking them."""

        # Check lamp status.
        command_off: bool = False
        lamp_status = self.helpers.lamps.list_status()
        for name, ls in lamp_status.items():
            if ls[1] is not False:
                self.command.warning(f"Lamp {name} is on, will turn off.")
                command_off = True

        if command_off:
            task = asyncio.create_task(self.helpers.lamps.all_off(self.command))
            if wait:
                await task

    async def _ensure_lamps(self, mode: str):
        """Ensures the lamps for flats/arcs/hartmann are on."""

        # Make sure FFS are closed.
        close_ffs = asyncio.create_task(self._close_ffs())

        # Check lamps. Depending on the other stages HgCd may be on but Ne not. Loop
        # over each on of the lamps and if it's not on, turn it on. If any of them is
        # not on wait for 10 seconds, which is enough for the Hartmanns.
        lamp_status = self.helpers.lamps.list_status()

        if mode == "flat":
            lamp_status = self.helpers.lamps.list_status()
            if lamp_status["ff"][3] is False:
                if self._lamps_task and not self._lamps_task.done():
                    # Lamps have been commanded on but are not warmed up yet.
                    await self._lamps_task
                else:
                    self.command.warning("Turning FF lamp on.")
                    await self.helpers.lamps.turn_lamp(
                        self.command,
                        ["ff"],
                        True,
                        turn_off_others=True,
                    )

        else:
            wait: float = 0
            for lamp in ["HgCd", "Ne"]:
                if lamp_status[lamp][0] is False:
                    self.command.warning(f"Turning {lamp} lamp on.")
                    asyncio.create_task(
                        self.helpers.lamps.turn_lamp(
                            self.command,
                            [lamp],
                            True,
                            turn_off_others=False,
                        )
                    )

                    # For hartmann we don't need to wait until the lamps have fully
                    # warmed up. For arcs we wait until they are.
                    if mode == "hartmann":
                        if lamp == "HgCd":
                            wait = 10 if wait < 10 else wait
                        else:
                            wait = 5 if wait < 5 else wait
                    else:
                        if LampsHelper.WARMUP[lamp] > wait:
                            wait = LampsHelper.WARMUP[lamp]

                elif mode == "arcs" and lamp_status[lamp][3] is False:
                    elapsed = lamp_status[lamp][2]
                    wait_lamp = LampsHelper.WARMUP[lamp] - elapsed
                    if wait < wait_lamp:
                        wait = wait_lamp

            if wait > 0:
                # TODO: maybe here we should confirm that the lamps are turning on.
                # We don't want to await the tasks though, because in some cases we
                # are waiting less time than the full warm up time.
                self.command.info(f"Waiting {wait} seconds for the lamps to warm-up.")
                await asyncio.sleep(wait)

        # Ensure FFS have fully closed.
        await close_ffs

    async def _set_guider_offset(self):
        """Sets the guider offset."""

        offset = self.config["guider_offset"]
        if offset is not None:
            offset = " ".join(map(str, offset))
            self.command.info(f"Setting guide offset to {offset}.")
            await self.send_command("cherno", f"offset {offset}")

    def _is_goto_complete(self):
        """Determines whether we can mark the configuration as ``goto_complete```.

        The stage at which we mark the configuration as ``goto_complete`` depends
        on what stages we are running, but it is always after the last stage
        before acquisition.

        """

        for stage in self.flat_stages:
            if stage == "acquire" or stage == "guide":
                continue
            if not self.is_stage_done(stage):
                return False

        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-16
# @Filename: apogee.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import enum

from typing import TYPE_CHECKING

from hal import config
from hal.exceptions import HALError

from . import HALHelper


if TYPE_CHECKING:
    from hal.actor import HALActor, HALCommandType


__all__ = ["APOGEEHelper"]


class APOGEEHelper(HALHelper):
    """APOGEE instrument helper."""

    name = "apogee"

    def __init__(self, actor: HALActor):
        super().__init__(actor)

        self.gang_helper = APOGEEGangHelper(actor)

    async def shutter(
        self,
        command: HALCommandType,
        open: bool = True,
        shutter: str = "apogee",
        force: bool = False,
    ):
        """Opens/closes the shutter.

        Parameters
        ----------
        command
            The command instance to use to command the shutter.
        open
            If `True`, opens the shutter, otherwise closes it.
        shutter
            The shutter to query. Can be ``apogee``, ``calbox``, or ``fpi``.
        force
            If `True`, sends the command to the shutter even if it reports to
            already be in that position.

        Returns
        -------
        shutter_command
            The command sent to the shutter after been awaited for completion or
            `None` if the shutter is already at that position.

        """

        position = "open" if open is True else "close"

        if force is False:
            current_position = self.get_shutter_position(shutter)
            if current_position is None and current_position == open:
                return None

        if shutter == "apogee":
            shutter_command = await self._send_command(
                command,
                "apogee",
                f"shutter {position}",
                time_limit=config["timeouts"]["apogee_shutter"],
            )

        elif shutter == "fpi":
            shutter_command = await self._send_command(
                command,
                "apogeefpi",
                position,
                time_limit=config["timeouts"]["apogee_shutter"],
            )

        elif shutter == "calbox":
            shutter_command = await self._send_command(
                command,
                "apogeecal",
                "shutterOpen" if open else "shutterClose",
                time_limit=config["timeouts"]["apogee_shutter"],
            )

        else:
            raise ValueError(f"Invalid shutter {shutter}.")

        return shutter_command

    def get_shutter_position(self, shutter: str = "apogee") -> bool | None:
        """Returns the shutter status.

        Parameters
        ----------
        shutter
            The shutter to query. Can be ``apogee``, ``calbox``, or ``fpi``.

        Returns
        -------
        shutter_status
            `True` if the shutter is open, `False` if closed, `None` if unknown.

        """

        shutter = shutter.lower()

        if shutter == "apogee":
            limit_switch = self.actor.models["apogee"]["shutterLimitSwitch"]
            if limit_switch is None or None in limit_switch.value:
                return None
            if limit_switch.value[0] is False and limit_switch.value[1] is True:
                return False
            elif limit_switch.value[1] is False and limit_switch.value[0] is True:
                return True
            else:
                return None

        elif shutter == "fpi":
            shutter_position = self.actor.models["apogeefpi"]["shutter_position"]

            if (
                shutter_position is None
                or shutter_position.value[0] is None
                or shutter_position.value[0] == "?"
            ):
                return None
            elif shutter_position.value[0].lower() == "closed":
                return False
            elif shutter_position.value[0].lower() == "open":
                return True
            else:
                return None

        elif shutter == "calbox":
            shutter_position = self.actor.models["apogeecal"]["calShutter"]

            if (
                shutter_position is None
                or shutter_position.value[0] is None
                or shutter_position.value[0] == "?"
            ):
                return None
            else:
                return shutter_position.value[0]
        else:
            raise ValueError(f"Invalid shutter {shutter}.")

    def get_dither_position(self) -> str | None:
        """Returns the dither position or `None` if unknown."""

        position = self.actor.models["apogee"]["ditherPosition"]
        if position is None or None in position.value:
            return None

        return position.value[1]

    async def set_dither_position(
        self,
        command: HALCommandType,
        position: str,
        force: bool = False,
    ):
        """Sets the dither mechanism to the commanded position."""

        position = position.upper()
        if position not in ["A", "B"]:
            raise HALError(f"Invalid dither position {position}.")

        current_position = self.get_dither_position()
        if current_position is None:
            command.warning("Current dither position is unknown.")

        if current_position == position and force is False:
            return None

        dither_command = await self._send_command(
            command,
            "apogee",
            f"dither namedpos={position}",
            time_limit=config["timeouts"]["apogee_dither"],
        )

        return dither_command

    def is_exposing(self):
        """Returns `True` if APOGEE is exposing or stopping."""

        exposure_state = self.actor.models["apogee"]["exposureState"]

        if exposure_state.value is None or None in exposure_state.value:
            raise ValueError("Unknown APOGEE exposure state.")

        state = exposure_state.value[0].lower()
        if state in ["exposing", "stopping"]:
            return True
        else:
            return False

    async def expose(
        self,
        command: HALCommandType,
        exp_time: float,
        exp_type: str = "dark",
        dither_position: str | None = None,
    ):
        """Exposes APOGEE.

        Parameters
        ----------
        command
            The command used to interact with the APOGEE actor.
        exp_time
            The exposure time.
        exp_type
            The exposure type. Valid values are ``object``, ``dark``, ``flat``,
            and ``DomeFlat``.
        dither_position
            The dither position. If `None`, uses the current position.

        """

        if exp_type.lower() not in ["object", "dark", "flat", "domeflat"]:
            raise HALError(f"Invalid exposure type {exp_type}.")

        if dither_position:
            await self.set_dither_position(command, dither_position)

        expose_command = await self._send_command(
            command,
            "apogee",
            f"expose time={exp_time:.1f} object={exp_type.lower()}",
            time_limit=exp_time + config["timeouts"]["expose"],
        )

        return expose_command

    async def expose_dither_pair(
        self,
        command: HALCommandType,
        exp_time: float,
        dither_sequence: str | None = None,
        exp_type: str = "object",
    ):
        """Takes an APOGEE dither set.

        Parameters
        ----------
        command
            The command used to interact with the APOGEE actor.
        exp_time
            The exposure time for each exposure in the dither sequence.
        exp_type
            The exposure type. Valid values are ``object``, ``dark``, and ``flat``.
        dither_sequence
            The dither sequence. If `None` the first dither will be taken at the
            current position and the mechanism will be switched after it.
            Alternatively, a string ``"AB"``, ``"BA"``, etc.

        """

        if dither_sequence is None:
            current = self.get_dither_position()
            if current is None:
                raise HALError("Cannot determine current APOGEE dither position.")
            dither_sequence = current.upper()
            dither_sequence = "AB" if dither_sequence == "A" else "BA"
        else:
            dither_sequence = dither_sequence.upper()
            if dither_sequence not in ["AB", "BA", "AA", "BB"]:
                raise HALError(f"Invalid dither sequence {dither_sequence}.")

        for dither_position in dither_sequence:
            await self.expose(
                command,
                exp_time,
                exp_type=exp_type,
                dither_position=dither_position,
            )


class APOGEEGangHelper:
    """Helper for the APOGEE gang connector."""

    def __init__(self, actor: HALActor):
        self.actor = actor
        self.flag: APOGEEGang = APOGEEGang.UNKNWON

        if self.actor.observatory == "APO":
            actor.models["mcp"]["apogeeGang"].register_callback(self._update_flag)

    async def _update_flag(self, value: list):
        """Callback to update the gang connector flag."""

        value = value or [0]
        self.flag = APOGEEGang(int(value[0]))

    def get_position(self) -> APOGEEGang:
        """Return the position of the gang connector."""

        return self.flag

    def at_podium(self):
        """Return True if the gang connector is on the podium."""

        ok = (self.get_position() & APOGEEGang.AT_PODIUM).value > 0
        return ok

    def at_cartridge(self):
        """Returns True if the gang connector is at the cartridge."""

        pos = self.get_position()
        ok = (pos == APOGEEGang.DISCONNECTED_FPI) or (pos == APOGEEGang.AT_FPS_FPI)
        return ok


class APOGEEGang(enum.Flag):
    """Flags for the ``mcp.apogeeGang`` keyword."""

    UNKNWON = 0
    DISCONNECTED = 1
    AT_CART = 2
    AT_PODIUM = 4
    PODIUM_DENSE = 12
    DISCONNECTED_FPI = 17
    AT_FPS_FPI = 18
    AT_PODIUM_SPARSE = 20
    AT_PODIUM_DENSE_FPI = 28
    AT_PODIUM_ONEM = 36
    AT_PODIUM_ONEM_FPI = 52



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-12-25
# @Filename: cherno.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
from enum import Flag
from time import time

from typing import TYPE_CHECKING

from clu.legacy.tron import TronKey

from hal.helpers import HALHelper


if TYPE_CHECKING:
    from hal.actor import HALActor, HALCommandType


__all__ = ["ChernoHelper", "GuiderStatus"]


class GuiderStatus(Flag):
    """Maskbits with the guider status."""

    IDLE = 1 << 0
    EXPOSING = 1 << 1
    PROCESSING = 1 << 2
    CORRECTING = 1 << 3
    STOPPING = 1 << 4
    FAILED = 1 << 5
    WAITING = 1 << 6
    UNKNOWN = 1 << 10

    NON_IDLE = EXPOSING | PROCESSING | CORRECTING | STOPPING | WAITING | UNKNOWN


class ChernoHelper(HALHelper):
    """Helper to interact with cherno."""

    name = "cherno"

    def __init__(self, actor: HALActor):
        super().__init__(actor)

        self.status = GuiderStatus.UNKNOWN

        self.model = actor.models["cherno"]
        self.model["guider_status"].register_callback(self._guider_status)

    async def _guider_status(self, key: TronKey):
        """Updates the internal guider status."""

        self.status = GuiderStatus(int(key.value[0], 16))

        if self.status.value == 0:
            self.status = GuiderStatus.UNKNOWN

    def is_guiding(self):
        """Returns `True` if the guider is not idle."""

        return bool(self.status & GuiderStatus.NON_IDLE)

    def guiding_at_rms(self, rms: float, max_age: float = 120):
        """Checks that the guider has reached a given RMS."""

        if not self.is_guiding():
            return False

        guide_rms = self.model["guide_rms"]

        if guide_rms and guide_rms.value and guide_rms.value[3]:
            last_seen = guide_rms.last_seen
            age = time() - last_seen
            if age > max_age:
                return False

            return guide_rms.value[3] <= rms
        else:
            return False

    async def wait_for_rms(self, rms: float, max_wait: float | None = None):
        """Blocks until a given RMS is reached."""

        elapsed = 0.0
        while True:
            if self.guiding_at_rms(rms):
                return True
            await asyncio.sleep(5)
            elapsed += 5
            if max_wait and elapsed > max_wait:
                raise asyncio.TimeoutError()

    async def acquire(
        self,
        command: HALCommandType,
        exposure_time: float = 15,
        target_rms: float | None = None,
        wait: bool | None = None,
        max_iterations: int | None = None,
    ):
        """Runs the acquisition command.

        Parameters
        ----------
        command
            The command to use to send the acquisition command.
        exposure_time
            The exposure time to use.
        target_rms
            The RMS to be reached by the acquisition command. If not specified,
            acquisition will be run in continuous mode.
        wait
            Whether to wait for the acquisition to be complete. If `None`, it will
            block only if ``target_rms`` has been defined.
        max_iterations
            Maximum number of iterations before failing.

        """

        block = True if (target_rms and not wait) or wait is True else False

        command_str = "acquire -c" if target_rms is None else f"acquire -r {target_rms}"
        command_str += f" -t {exposure_time}"

        if max_iterations:
            command_str += f" -x {max_iterations}"

        if block:
            await self._send_command(command, "cherno", command_str)
        else:
            asyncio.create_task(self._send_command(command, "cherno", command_str))

        return

    async def guide(
        self,
        command: HALCommandType,
        exposure_time: float = 15,
        wait: bool = False,
    ):
        """Start the guide loop."""

        coro = self._send_command(command, "cherno", f"guide -t {exposure_time}")

        if wait:
            await coro
        else:
            asyncio.shield(asyncio.create_task(coro))

    async def stop_guiding(self, command: HALCommandType):
        """Stops the guide loop."""

        await self._send_command(command, "cherno", "stop")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-09-26
# @Filename: goto.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

from hal.exceptions import HALError

from . import hal_command_parser


if TYPE_CHECKING:
    from . import HALCommandType


__all__ = ["gotoStow"]


async def goto_position(command: HALCommandType, name: str):
    """Go to position."""

    try:
        await command.actor.helpers.tcc.goto_position(command, name)
    except HALError as err:
        return command.fail(f"Goto position failed with error: {err}")

    return command.finish()


@hal_command_parser.command(name="gotoStow", cancellable=True)
async def gotoStow(command: HALCommandType):
    """Send the telescope to (120, 30, 0)."""

    return await goto_position(command, "stow")


@hal_command_parser.command(name="gotoAll60", cancellable=True)
async def gotoAll60(command: HALCommandType):
    """Send the telescope to (60, 60, 60)."""

    return await goto_position(command, "all_60")


@hal_command_parser.command(name="gotoStow60", cancellable=True)
async def gotoStow60(command: HALCommandType):
    """Send the telescope to (121, 60, 0)."""

    return await goto_position(command, "stow_60")


@hal_command_parser.command(name="gotoInstrumentChange", cancellable=True)
async def gotoInstrumentChange(command: HALCommandType):
    """Send the telescope to (121, 90, 0)."""

    return await goto_position(command, "instrument_change")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-03-24
# @Filename: exceptions.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)


class HALError(Exception):
    """A custom core HAL exception"""

    def __init__(self, message=None):
        message = "There has been an error" if not message else message

        super(HALError, self).__init__(message)


class MacroError(HALError):
    """An error during a macro execution."""

    pass


class HALNotImplemented(HALError):
    """A custom exception for not yet implemented features."""

    def __init__(self, message=None):
        message = "This feature is not implemented yet." if not message else message

        super(HALNotImplemented, self).__init__(message)


class HALMissingDependency(HALError):
    """A custom exception for missing dependencies."""

    pass


class HALWarning(Warning):
    """Base warning for HAL."""


class HALUserWarning(UserWarning, HALWarning):
    """The primary warning class."""

    pass


class HALDeprecationWarning(HALUserWarning):
    """A warning for deprecated features."""

    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-09-26
# @Filename: tcc.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

from hal import config
from hal.exceptions import HALError

from . import HALHelper


if TYPE_CHECKING:
    from hal.actor import HALCommandType


__all__ = ["TCCHelper"]


class TCCHelper(HALHelper):
    """Helper for the TCC."""

    name = "tcc"

    is_slewing: bool = False

    async def goto_position(
        self,
        command: HALCommandType,
        where: str | dict[str, float],
        **kwargs,
    ):
        """Executes the goto command.

        Parameters
        ----------
        command
            The actor command.
        where
            The name of the goto entry in the configuration file, or a
            dictionary with the keys ``(alt, az, rot)`` in degrees.
        kwargs
            Arguments to pass to `~.TCCHelper.do_slew`.

        """

        config = command.actor.config

        if isinstance(where, str):
            if where not in config["goto"]:
                raise HALError(f"Cannot find goto position '{where}'.")

            alt = config["goto"][where]["alt"]
            az = config["goto"][where]["az"]
            rot = config["goto"][where]["rot"]

            where = {"alt": alt, "az": az, "rot": rot}

        if self.is_slewing:
            raise HALError("TCC is already slewing.")

        # await self.axis_init(command)

        # Even if this is already checked in axis_init(), let's check again that the
        # axes are ok, but if alt < limit, we only check az and alt because we won't
        # move in altitude.
        result = self.axes_are_clear()
        if not result:
            raise HALError("Some axes are not clear. Cannot continue.")

        # Now do the actual slewing.
        slew_result = await self.do_slew(command, where, **kwargs)
        if slew_result is False:
            raise HALError(f"Failed going to position {where}.")

        return command.info(text=f"At position {where}.")

    async def axis_init(self, command: HALCommandType) -> bool:
        """Executes TCC axis init or fails."""

        status = await self._send_command(
            command,
            "tcc",
            "axis status",
            time_limit=20.0,
            raise_on_fail=False,
        )
        if status.status.did_fail:
            raise HALError("'tcc status' failed. Is the TCC connected?")

        if self.check_stop_in() is True:
            raise HALError(
                "Cannot tcc axis init because of bad axis status: "
                "Check stop buttons on Interlocks panel."
            )

        sem = await self.mcp_semaphore_ok(command)
        if sem is False:
            raise HALError("Failed getting the semaphore information.")

        if sem == "TCC:0:0" and self.axes_are_clear():
            command.debug(
                text="Axes clear and TCC has semaphore. "
                "No axis init needed, so none sent."
            )
            return True

        command.debug(text="Sending tcc axis init.")
        axis_init_cmd_str = "axis init"
        if self.below_alt_limit():
            command.warning(
                text="Altitude below interlock limit! Only initializing altitude "
                "and rotator: cannot move in az."
            )
            axis_init_cmd_str += " rot,alt"
        axis_init_cmd = await self._send_command(
            command,
            "tcc",
            axis_init_cmd_str,
            time_limit=20.0,
            raise_on_fail=False,
        )

        if axis_init_cmd.status.did_fail:
            command.error("Cannot slew telescope: failed tcc axis init.")
            raise HALError("Cannot slew telescope: check and clear interlocks?")

        return True

    async def axis_stop(self, command: HALCommandType, axis: str = "") -> bool:
        """Issues an axis stop to the TCC."""

        axis_stop_cmd = await self._send_command(
            command,
            "tcc",
            f"axis stop {axis}".strip(),
            time_limit=30.0,
            raise_on_fail=False,
        )

        if axis_stop_cmd.status.did_fail:
            raise HALError("Error: failed to cleanly stop telescope via tcc axis stop.")

        self.is_slewing = False
        return True

    def get_bad_axis_bits(self, axes=("az", "alt", "rot"), mask=None):
        """Return the bad status bits for the requested axes."""

        if mask is None:
            mask = self.actor.models["tcc"]["axisBadStatusMask"][0]

        if mask is None:
            self.actor.write("e", text="Cannot retrieve TCC axisBadStatusMask.")
            return [1, 1, 1]

        return [(self.actor.models["tcc"][f"{axis}Stat"][3] & mask) for axis in axes]

    def check_stop_in(self, axes=("az", "alt", "rot")) -> bool:
        """Returns `True` if any stop bit is set in the ``<axis>Stat`` TCC keywords.

        The [az,alt,rot]Stat[3] bits show the exact status:
        http://www.apo.nmsu.edu/Telescopes/HardwareControllers/AxisControllers.html#25mStatusBits

        """
        try:
            # 0x2000 is "stop button in"
            return any(self.get_bad_axis_bits(axes=axes, mask=0x2000))
        except TypeError:
            # Some axisStat is unknown (and thus None)
            return False

    async def mcp_semaphore_ok(self, command: HALCommandType):
        """Returns the semaphore if the semaphore is owned by the TCC or nobody."""

        mcp_model = self.actor.models["mcp"]

        sem = mcp_model["semaphoreOwner"]
        if sem is None:
            sem_show_cmd = await self._send_command(
                command,
                "mcp",
                "sem.show",
                time_limit=20.0,
                raise_on_fail=False,
            )

            if sem_show_cmd.status.did_fail:
                raise HALError("Cannot get mcp semaphore. Is the MCP alive?")

            sem = mcp_model["semaphoreOwner"]

        if (
            (sem[0] != "TCC:0:0")
            and (sem[0] != "")
            and (sem[0] != "None")
            and (sem[0] is not None)
        ):
            raise HALError(
                f"Cannot axis init: Semaphore is owned by  {sem[0]}. "
                "If you are the owner (e.g., via MCP Menu), release it and try again. "
                "If you are not the owner, confirm that you can steal "
                "it from them, then issue: mcp sem.steal"
            )

        return sem

    def axes_are_clear(self, axes=("az", "alt", "rot")) -> bool:
        """Checks that no bits are set in any axis status field."""

        if "tcc" in self.actor.helpers.bypasses:
            self.actor.write(
                "w",
                text="Saying all axes are clear because TCC is bypassed.",
            )
            return True

        axes = [ax for ax in axes if ax != "alt" or self.below_alt_limit() is False]
        try:
            tcc_model = self.actor.models["tcc"]
            return all((tcc_model[f"{axis}Stat"][3] == 0) for axis in axes)
        except TypeError:
            # Some axisStat is unknown (and thus None)
            return False

    def below_alt_limit(self) -> bool:
        """Check if we are below the alt=18 limit that prevents init/motion in az."""

        limit = self.actor.config["goto"]["alt_limit"]
        return self.actor.models["tcc"]["axePos"][1] < limit

    def check_axes_status(self, status: str) -> bool:
        """Returns `True` if all the axes are at ``status``."""

        axes_status = self.actor.models["tcc"]["AxisCmdState"].value
        return all([axis.lower() == status.lower() for axis in axes_status])

    async def do_slew(
        self,
        command,
        coords: dict[str, float] | None = None,
        track_command: str | None = None,
        keep_offsets: bool = True,
        offset: bool = False,
        rotwrap: str | None = None,
    ) -> bool:
        """Correctly handle a slew command, given what parse_args had received.

        Parameters
        ----------
        command
            The actor command used to send child commands to the TCC.
        coords
            The coordinates where to slew. It must be a dictionary with keys
            ``ra, dec, rot`` or ``alt, az, rot``.
        track_command
            A raw TCC ``track`` command (without the ``tcc`` target) to send.
            In this case other arguments like ``keep_offsets`` or ``rotwrap``
            are ignored.
        keep_offsets
            Whether to keep the existing offsets.
        rotwrap
            The type of ``/RotWrap`` to use.
        offset
            If defined, the coordinates will be treated as an offset and not
            absolute positions.

        """

        tcc_model = self.actor.models["tcc"]

        # Check axes.
        result = self.axes_are_clear()
        if not result:
            raise HALError("Some axes are not clear. Cannot continue.")

        # NOTE: TBD: We should limit which offsets are kept.
        keep_args = "/keep=(arc,gcorr,calib,bore)" if keep_offsets else ""
        rotwrap = f"/rotwrap={rotwrap}" if rotwrap else ""

        slew_cmd = None
        if track_command:
            slew_cmd = self._send_command(
                command,
                "tcc",
                track_command,
                time_limit=config["timeouts"]["slew"],
                raise_on_fail=False,
            )
        else:
            if coords is None:
                raise HALError("coords needed if track_command is None.")

            if not offset:
                if "ra" in coords and "dec" in coords and "rot" in coords:
                    ra = coords["ra"]
                    dec = coords["dec"]
                    rot = coords["rot"]

                    command.info(
                        text="Slewing to (ra, dec, rot) == "
                        f"({ra:.4f}, {dec:.4f}, {rot:g})"
                    )
                    if keep_args:
                        command.warning(text="keeping all offsets")

                    slew_cmd = self._send_command(
                        command,
                        "tcc",
                        f"track {ra}, {dec} icrs /rottype=object/rotang={rot:g} "
                        f"{rotwrap} {keep_args}",
                        time_limit=config["timeouts"]["slew"],
                        raise_on_fail=False,
                    )

                elif "az" in coords and "alt" in coords and "rot" in coords:
                    alt = coords["alt"]
                    az = coords["az"]
                    rot = coords["rot"]

                    command.info(
                        text="Slewing to (az, alt, rot) == "
                        f"({az:.4f}, {alt:.4f}, {rot:.4f})"
                    )

                    slew_cmd = self._send_command(
                        command,
                        "tcc",
                        f"track {az:f}, {alt:f} mount /rottype=mount "
                        f"/rotangle={rot:f} {rotwrap} {keep_args}",
                        time_limit=config["timeouts"]["slew"],
                        raise_on_fail=False,
                    )

                else:
                    raise HALError("Not enough coordinates information provided.")

            else:
                if "alt" not in coords or "az" not in coords:
                    raise HALError("No alt/az offsets provided.")

                # In arcsec
                alt = coords["alt"] or 0.0
                az = coords["az"] or 0.0
                rot = coords["rot"] or 0.0

                command.info(text=f"Offseting alt={alt:.3f}, az={az:.3f}")

                slew_cmd = self._send_command(
                    command,
                    "tcc",
                    f"offset guide {az/3600.:g},{alt/3600.:g},{rot/3600.:g} /computed",
                    time_limit=config["timeouts"]["slew"],
                    raise_on_fail=False,
                )

        # "tcc track" in the new TCC is only Done successfully when all requested
        # axes are in the "tracking" state. All other conditions mean the command
        # failed, and the appropriate axisCmdState and axisErrCode will be set.
        # However, if an axis becomes bad during the slew, the TCC will try to
        # finish it anyway, so we need to explicitly check for bad bits.

        try:
            self.is_slewing = True
            slew_cmd = await slew_cmd
        finally:
            self.is_slewing = False

        if slew_cmd.status.did_fail:
            str_axis_state = ",".join(tcc_model["axisCmdState"].value)
            str_axis_code = ",".join(tcc_model["axisErrCode"].value)
            command.error(
                f"tcc track command failed with axis states: {str_axis_state} "
                f"and error codes: {str_axis_code}"
            )
            raise HALError("Failed to complete slew: see TCC messages for details.")

        if self.axes_are_clear() is False:
            axis_bits = self.get_bad_axis_bits()
            command.error(
                "TCC slew command ended with some bad bits set: "
                "0x{:x},0x{:x},0x{:x}".format(*axis_bits)
            )
            raise HALError("Failed to complete slew: see TCC messages for details.")

        return True



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-10-10
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

import glob
import importlib
import inspect
import os
import warnings

from hal.exceptions import HALUserWarning

from .macro import Macro


# Dynamically inspect all the files in this directory and import the subclasses
# of Macro. Also create a list of instances for the macros.

# TODO: maybe Macro subclasses should be a singleton ...


exclusions = ["__init__.py", "macro.py"]

cwd = os.getcwd()
os.chdir(os.path.dirname(os.path.realpath(__file__)))

files = [f_ for f_ in glob.glob("**/*.py", recursive=True) if f_ not in exclusions]
all_macros: list[Macro] = []

for f_ in files:
    try:
        modname = f_[0:-3].replace("/", ".")
        mod = importlib.import_module("hal.macros." + modname)
        for objname in dir(mod):
            obj = getattr(mod, objname)
            if inspect.isclass(obj) and issubclass(obj, Macro) and obj != Macro:
                all_macros.append(obj())
                locals().update({objname: obj})

    except Exception as ee:
        warnings.warn(f"cannot import file {f_}: {ee}", HALUserWarning)

os.chdir(cwd)

__all__ = ["all_macros", "Macro"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-12-26
# @Filename: auto.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from hal.macros.expose import ExposeMacro

from . import hal_command_parser


if TYPE_CHECKING:
    from .. import HALCommandType


__all__ = ["auto"]


@hal_command_parser.command(name="auto")
@click.option(
    "--stop",
    is_flag=True,
    help="Stops the auto mode loop after the next stage completes. "
    "For an immediate stop use with --now.",
)
@click.option(
    "--now",
    is_flag=True,
    help="Along with --stop, cancels the auto mode loop immediately.",
)
@click.option(
    "--modify",
    is_flag=True,
    help="Modify an ongoing auto loop.",
)
@click.option(
    "--pause",
    is_flag=True,
    help="Pauses the execution of the macro. The current exposures will complete.",
)
@click.option(
    "--resume",
    is_flag=True,
    help="Resumes the execution of the macro.",
)
@click.option(
    "--count",
    type=int,
    default=1,
    help="Number of exposures per design.",
)
async def auto(
    command: HALCommandType,
    stop: bool = False,
    now: bool = False,
    modify: bool = False,
    pause: bool = False,
    resume: bool = False,
    count: int = 1,
):
    """Starts the auto mode."""

    assert command.actor

    expose_macro = command.actor.helpers.macros["expose"]
    assert isinstance(expose_macro, ExposeMacro)

    macro = command.actor.helpers.macros["auto"]

    if (stop or modify or pause or resume) and not macro.running:
        return command.fail(
            "I'm afraid I cannot do that Dave. The auto mode is not running."
        )

    if pause and resume:
        return command.fail("--pause and --resume are incompatible Dave.")

    if pause:
        await expose_macro._pause()
        return command.finish()

    if resume:
        await expose_macro._resume()
        return command.finish()

    if stop is True:
        if now is True:
            command.warning(auto_mode_message="Cancelling auto mode NOW.")
            macro.cancel(now=True)
        else:
            command.warning(auto_mode_message="Cancelling auto after stage completes.")
            macro.cancel(now=False)

        return command.finish()

    if modify is True:
        # For now the only option we can modify is the count of exposures.
        macro.config["count"] = count

        # Also modify active expose macro, if any is running.
        if expose_macro.running:
            expose_macro.expose_helper.update_params(
                count_boss=count,
                count_apogee=count,
            )
            expose_macro.expose_helper.refresh()

        return command.finish()

    # From this point on this is a new macro, so it should not be already running.
    if macro.running:
        return command.fail(
            "I'm afraid I cannot do that Dave. The auto mode is already running."
        )

    result: bool = True
    while True:
        # Run the auto loop until the command is cancelled.
        macro.reset(command, count=count)
        if not await macro.run():
            result = False
            break

        if macro.cancelled:
            # Cancelled macros return result=True
            break

    if result is False:
        return command.fail()

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-16
# @Filename: apogee_dome_flat.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from clu.legacy.tron import TronKey

from hal import config
from hal.exceptions import MacroError
from hal.macros import Macro


__all__ = ["APOGEEDomeFlatMacro"]


class APOGEEDomeFlatMacro(Macro):
    """Take an APOGEE dome flat after an exposure."""

    name = "apogee_dome_flat"

    __PRECONDITIONS__ = ["gang_at_cart"]
    __STAGES__ = [("ffs", "open_shutter"), "expose"]
    __CLEANUP__ = ["cleanup"]

    __ffs_initial_state: str | None = None

    async def gang_at_cart(self):
        """Checks that the gang connected is at the cart."""

        if not self.command.actor.helpers.apogee.gang_helper.at_cartridge():
            raise MacroError("The APOGEE gang connector is not at the cart.")

        return True

    async def ffs(self):
        """Check the FFS status and closes it."""

        if self.command.actor.helpers.ffs.all_closed():
            self.command.debug("FFS already closed.")
            self.__ffs_initial_state = "closed"
            return
        elif self.command.actor.helpers.ffs.all_open():
            self.__ffs_initial_state = "open"
        else:
            raise MacroError("Cannot determine state of FFS.")

        self.command.info("Closing FFS.")

        await self.command.actor.helpers.ffs.close(self.command)

    async def open_shutter(self):
        """Opens the APOGEE cold shutter."""

        apogee = self.command.actor.helpers.apogee

        if not await apogee.shutter(self.command, True):
            raise MacroError("Failed opening APOGEE shutter.")

    async def expose(self):
        """Takes the dome flat. Turns on the FFS after the fourth read."""

        apogee = self.command.actor.helpers.apogee

        self.command.actor.models["apogee"]["utrReadState"].register_callback(
            self._flash_lamps
        )

        self.command.info("Taking APOGEE dome flat.")
        await apogee.expose(self.command, 50, exp_type="DomeFlat")

        return True

    async def _flash_lamps(self, key: TronKey):
        """Flashes the FF lamps."""

        if len(key.value) == 0:
            return

        state = key.value[1]
        n_read = int(key.value[2])
        if n_read == 3 and state == "Reading":
            time_to_flash = 4.0

            self.command.info(text="Calling ff_lamp.on")

            asyncio.create_task(
                self.send_command(
                    "mcp",
                    "ff.on",
                    time_limit=config["timeouts"]["lamps"],
                )
            )
            await asyncio.sleep(time_to_flash)

            lamp_off = await self.command.send_command("mcp", "ff.off")
            if lamp_off.status.did_fail:
                raise MacroError("Failed flashing lamps.")

        return

    async def cleanup(self):
        """Closes the shutter and does cleanup."""

        apogee = self.command.actor.helpers.apogee

        try:
            self.command.actor.models["apogee"]["utrReadState"].remove_callback(
                self._flash_lamps
            )
        except AssertionError:
            pass

        if not await apogee.shutter(self.command, False):
            self.command.error("Failed closing APOGEE shutter.")

        if self.__ffs_initial_state == "open":
            self.command.info("Reopening FFS.")
            await self.command.actor.helpers.ffs.open(self.command)
        else:
            self.command.info("Keeping FFS closed.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-20
# @Filename: lamps.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import time

from typing import TYPE_CHECKING

from hal import config
from hal.exceptions import HALError

from . import HALHelper


if TYPE_CHECKING:
    from hal.actor import HALCommandType


__all__ = ["LampsHelper"]


class LampsHelper(HALHelper):
    """Control for lamps."""

    LAMPS = ["ff", "HgCd", "Ne"]
    WARMUP = config["lamp_warmup"].copy()

    name = "lamps"

    def _command_one(self, command: HALCommandType, lamp: str, state: bool):
        """Commands one lamp."""

        state_str = "on" if state is True else "off"

        return self._send_command(
            command,
            "mcp",
            f"{lamp.lower()}_{state_str}",
            time_limit=config["timeouts"]["lamps"],
        )

    async def all_off(self, command: HALCommandType, force: bool = False):
        """Turn off all the lamps."""

        status = self.list_status()

        tasks = []
        for lamp in self.LAMPS:
            if force is False and status[lamp][0] is False and status[lamp][-1] is True:
                continue
            tasks.append(self.turn_lamp(command, lamp, False, force=force))

        await asyncio.gather(*tasks)

    def list_status(self) -> dict[str, tuple[bool, bool, float, bool]]:
        """Returns a dictionary with the state of the lamps.

        For each lamp the first value in the returned tuple is whether the
        lamp has been commanded on. The second value is whether the lamps is
        actually on. The third value is how long since it changed states.
        The final value indicates whether the lamp has warmed up.

        """

        state = {}
        for lamp in self.LAMPS:
            commanded_key = f"{lamp}LampCommandedOn"
            commanded_on = self.actor.models["mcp"][commanded_key][0]
            if commanded_on is None:
                raise HALError(f"Failed getting {commanded_key}.")
            if lamp in ["wht", "UV"]:
                is_on = bool(commanded_on)
                state[lamp] = (is_on, is_on, 0.0, is_on)
            else:
                lamp_key = f"{lamp}Lamp"
                lamp_state = self.actor.models["mcp"][lamp_key]
                if any([lv is None for lv in lamp_state]):
                    raise HALError(f"Failed getting {lamp_key}.")
                last_seen = lamp_state.last_seen
                if sum(lamp_state.value) == 4:
                    lamp_state = True
                else:  # Sometimes when a lamp is turning on we'll have, e.g., 1,0,0,1.
                    lamp_state = False

                elapsed = time.time() - last_seen
                warmed = (elapsed >= self.WARMUP[lamp]) if bool(commanded_on) else False
                state[lamp] = (bool(commanded_on), lamp_state, elapsed, warmed)

        return state

    async def turn_lamp(
        self,
        command: HALCommandType,
        lamps: str | list[str],
        state: bool,
        turn_off_others: bool = False,
        delay: float = 0.0,
        force: bool = False,
    ):
        """Turns a lamp on or off.

        This routine always blocks until the lamps are on and warmed up. If
        you don't want to block, call it as a task.

        Parameters
        ----------
        command
            The command that issues the lamp switching.
        lamps
            Name of the lamp(s) to turn on or off.
        state
            Whether to turn lamp on or off.
        turn_off_others
            Turn off all other lamps.
        delay
            Wait this amount of seconds before actually commanding the lamps.
        force
            If `True`, send the on/off command regardless of status.

        """

        if isinstance(lamps, str):
            lamps = [lamps]

        for lamp in lamps:
            if lamp not in self.LAMPS:
                raise HALError(f"Invalid lamp {lamp}.")

        await asyncio.sleep(delay)

        status = self.list_status()

        tasks = []
        turn_off_tasks = []
        for ll in self.LAMPS:
            if ll in lamps:
                if force is False and status[ll][0] is state and status[ll][-1] is True:
                    pass
                else:
                    tasks.append(self._command_one(command, ll, state))
            else:
                if turn_off_others is True:
                    if status[ll][0] is not False or force is True:
                        turn_off_tasks.append(self._command_one(command, ll, False))

        if len(turn_off_tasks) > 0:
            await asyncio.gather(*turn_off_tasks)

        await asyncio.gather(*tasks)

        done_lamps: list[bool] = [False] * len(lamps)
        warmed: list[bool] = [False] * len(lamps)

        n_iter = 0
        while True:
            if all(done_lamps):
                if state is False:
                    if len(lamps) > 0:
                        command.info(f"Lamps {','.join(lamps)} are off.")
                    else:
                        command.info(f"Lamp {','.join(lamps)} is off.")
                    return
                elif all(warmed):
                    if len(lamps) > 0:
                        command.info(f"Lamps {','.join(lamps)} are on and warmed up.")
                    else:
                        command.info(f"Lamp {','.join(lamps)} is on and warmed up.")
                    return

            new_status = self.list_status()
            for i, lamp in enumerate(lamps):
                # Don't don't anything if we're already at the state.
                if done_lamps[i] and (state is False or warmed[i]):
                    continue

                # Index 1 is if the lamp is actually on/off, not only commanded.
                if new_status[lamp][1] is state:
                    done_lamps[i] = True

                if state is True:
                    # Index 2 is the elapsed time since it was completely on/off.
                    elapsed = new_status[lamp][2]
                    if elapsed >= self.WARMUP[lamp]:
                        command.debug(f"Lamp {lamp}: warm-up complete.")
                        warmed[i] = True
                    elif (n_iter % 5) == 0:
                        remaining = int(self.WARMUP[lamp] - elapsed)
                        command.debug(
                            f"Warming up lamp {lamp}: " f"{remaining} s remaining."
                        )

            await asyncio.sleep(1)
            n_iter += 1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-16
# @Filename: calibrations.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

from . import hal_command_parser, stages


if TYPE_CHECKING:
    from hal.macros import Macro

    from .. import HALCommandType


@hal_command_parser.group()
def calibrations():
    """Performs camera and telescope calibrations."""

    pass


@calibrations.command(name="apogee-dome-flat")
@stages("apogee_dome_flat")
async def apogee_dome_flat(command: HALCommandType, macro: Macro):
    """Runs the APOGEE dome flat sequence."""

    result = await macro.run()

    if result is False:
        return command.fail()

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-03-24
# @Filename: __main__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import os
import sys

import click
from click_default_group import DefaultGroup

from clu.tools import cli_coro
from sdsstools.daemonizer import DaemonGroup

from hal import __version__
from hal.actor import HALActor


@click.group(
    cls=DefaultGroup,
    default="actor",
    default_if_no_args=True,
    invoke_without_command=True,
)
@click.option(
    "-c",
    "--config",
    "config_file",
    type=click.Path(exists=True, dir_okay=False),
    help="Path to the user configuration file.",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    help="Debug mode.",
)
@click.option(
    "--version",
    is_flag=True,
    help="Print version and exit.",
)
@click.pass_context
def hal(
    ctx: click.Context,
    config_file: str | None = None,
    verbose: bool = False,
    version: bool = False,
):
    """HAL actor."""

    ctx.obj = {"verbose": verbose, "config_file": config_file}

    if version is True:
        click.echo(__version__)
        sys.exit(0)


@hal.group(cls=DaemonGroup, prog="hal-actor", workdir=os.getcwd())
@click.pass_context
@cli_coro
async def actor(ctx: click.Context):
    """Runs the actor."""

    default_config_file = os.path.join(os.path.dirname(__file__), "etc/hal.yml")
    config_file: str = ctx.obj["config_file"] or default_config_file

    hal_obj = HALActor.from_config(config_file)

    if ctx.obj["verbose"]:
        hal_obj.log.sh.setLevel(0)

    await hal_obj.start()
    await hal_obj.run_forever()


if __name__ == "__main__":
    hal()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-01-14
# @Filename: bypass.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from clu.parsers.click import CluGroup

from . import hal_command_parser


if TYPE_CHECKING:
    from .. import HALCommandType


__all__ = ["bypass"]


@hal_command_parser.group(cls=CluGroup)
def bypass():
    """Enable/disable bypasses."""

    pass


@bypass.command()
@click.argument("BYPASSES", type=str, nargs=-1)
async def enable(command: HALCommandType, bypasses: list[str]):
    """Enables a series of bypasses"""

    for bypass in bypasses:
        if bypass not in command.actor.helpers._available_bypasses:
            return command.fail(f"Invalid bypass name {bypass}.")

    for bypass in bypasses:
        command.actor.helpers.bypasses.add(bypass)

    return command.finish(bypasses=list(command.actor.helpers.bypasses))


@bypass.command()
@click.argument("BYPASSES", type=str, nargs=-1)
async def disable(command: HALCommandType, bypasses: list[str]):
    """Disabless a series of bypasses"""

    for bypass in bypasses:
        if bypass not in command.actor.helpers._available_bypasses:
            return command.fail(f"Invalid bypass name {bypass}.")

    for bypass in bypasses:
        if bypass in command.actor.helpers.bypasses:
            command.actor.helpers.bypasses.remove(bypass)

    return command.finish(bypasses=list(command.actor.helpers.bypasses))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-09-27
# @Filename: script.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from . import hal_command_parser


if TYPE_CHECKING:
    from . import HALCommandType


__all__ = ["script"]


@hal_command_parser.group()
def script():
    """Tools to list and run scripts."""

    pass


@script.command(name="list")
async def list_(command: HALCommandType):
    """Returns a list of available scripts."""

    available_scripts = command.actor.helpers.scripts.list_scripts()
    return command.finish(available_scripts=available_scripts)


@script.command()
async def running(command: HALCommandType):
    """Returns a list of running scripts."""

    running_scripts = list(command.actor.helpers.scripts.running.keys())
    return command.finish(running_scripts=running_scripts)


@script.command()
@click.argument("SCRIPT", type=str)
async def get_steps(command: HALCommandType, script: str):
    """Lists the steps in the script."""

    try:
        steps = command.actor.helpers.scripts.get_steps(script)
    except Exception as err:
        return command.fail(error=err)

    for step in steps:
        if step[2] is None:
            command.info(text=f"{step[0]} {step[1]}")
        else:
            command.info(text=f"{step[2]} {step[0]} {step[1]}")

    return command.finish()


@script.command()
@click.argument("SCRIPT", type=str)
async def run(command: HALCommandType, script: str):
    """Runs a script."""

    try:
        result = await command.actor.helpers.scripts.run(script, command)
    except Exception as err:
        return command.fail(error=err)

    if result is True:
        return command.finish(text=f"Script {script} has been successfully executed.")
    else:
        return command.fail(error=f"Script {script} failed or was cancelled.")


@script.command()
@click.argument("SCRIPT", type=str)
async def cancel(command: HALCommandType, script: str):
    """Cancels the execution of a script."""

    try:
        await command.actor.helpers.scripts.cancel(script)
    except Exception as err:
        command.warning(text=f"Error found while trying to cancel {script}.")
        return command.fail(error=err)

    return command.finish(f"Script {script} has been scheduled for cancellation.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-09-26
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING, cast

from clu import Command, CommandStatus

from hal.exceptions import HALError


if TYPE_CHECKING:
    from hal.actor import HALActor, HALCommandType


class HALHelper:
    """A helper class to control an actor or piece of hardware."""

    name: str | None = None

    def __init__(self, actor: HALActor):
        self.actor = actor

    async def _send_command(
        self,
        command: HALCommandType,
        target: str,
        cmd_str: str,
        raise_on_fail: bool = True,
        **kwargs,
    ):
        """Sends a command to a target."""

        bypasses = self.actor.helpers.bypasses

        # If the helper is bypassed, just returns a fake done command.
        if (self.name and self.name in bypasses) or ("all" in bypasses):
            command.warning(f"Bypassing command '{target} {cmd_str}'")
            cmd = Command()
            cmd.set_status(CommandStatus.DONE)
            return cmd

        if self.actor.tron is None or self.actor.tron.connected() is False:
            raise HALError("Not connected to Tron. Cannot send commands.")

        cmd = await command.send_command(target, cmd_str, **kwargs)
        if raise_on_fail and cmd.status.did_fail:
            if cmd.status == CommandStatus.TIMEDOUT:
                raise HALError(f"Command '{target} {cmd_str}' timed out.")
            else:
                raise HALError(f"Command '{target} {cmd_str}' failed.")

        return cast(Command, cmd)


from .apogee import *
from .boss import *
from .cherno import *
from .ffs import *
from .jaeger import *
from .lamps import *
from .scripts import *
from .tcc import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-10-10
# @Filename: goto_field.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING

import click

from hal import config

from . import hal_command_parser, stages


if TYPE_CHECKING:
    from hal.macros import Macro
    from hal.macros.macro import StageType

    from .. import HALCommandType


__all__ = ["goto_field"]


@hal_command_parser.command(name="goto-field", cancellable=True)
@stages("goto_field", reset=False)
@click.option(
    "--auto",
    is_flag=True,
    help="Selects the stages based on the latest design loaded.",
)
@click.option(
    "--guider-time",
    type=float,
    default=config["macros"]["goto_field"]["guider_time"],
    show_default=True,
    help="Exposure time for guiding/acquisition.",
)
@click.option(
    "--fixed-rot/--no-fixed-rot",
    default=False,
    help="Slews to a fixed rot position for the FVC loop. If --no-fixed-rot then "
    "--fixed-altaz is ignored.",
)
@click.option(
    "--fixed-altaz/--no-fixed-altaz",
    default=False,
    help="Slews to a fixed alt/az position for the FVC loop.",
)
@click.option(
    "--alt",
    type=float,
    help="The fixed altitude angle to which to slew for the FVC loop. "
    "Requires --fixed-altaz to take effect.",
)
@click.option(
    "--az",
    type=float,
    help="The fixed azimuth angle to which to slew for the FVC loop. "
    "Requires --fixed-altaz to take effect.",
)
@click.option(
    "--rot",
    type=float,
    help="The fixed rotator angle to which to slew for the FVC loop.",
)
@click.option(
    "--keep-offsets/--no-keep-offsets",
    is_flag=True,
    default=True,
    help="Keep the guider offsets from the previous field.",
)
@click.option(
    "--with-hartmann",
    is_flag=True,
    help="Ensures the boss_hartmann stage is selected. Mostly relevant with --auto.",
)
async def goto_field(
    command: HALCommandType,
    macro: Macro,
    stages: list[StageType],
    guider_time: float,
    auto: bool = False,
    fixed_rot: bool = False,
    fixed_altaz: bool = False,
    alt: float | None = None,
    az: float | None = None,
    rot: float | None = None,
    keep_offsets: bool = True,
    with_hartmann: bool = False,
):
    """Execute the go-to-field macro."""

    assert command.actor

    jaeger_helper = command.actor.helpers.jaeger

    if stages is not None and auto is True:
        return command.fail("--auto cannot be used with custom stages.")
    elif auto is True:
        configuration = jaeger_helper.configuration

        # The jaeger helper marks if a configuration is a new field or not but
        # only if the previous configuration has been observed, which only works
        # in full HAL auto mode. For goto-field auto mode, we can only check if the
        # previous configuration has the same field_id, regardless of whether it
        # was observed or not.
        previous = jaeger_helper._previous[-1] if jaeger_helper._previous else None

        if configuration is None:
            return command.fail("No configuration loaded. Auto mode cannot be used.")
        elif configuration.cloned is True:
            stages = config["macros"]["goto_field"]["cloned_stages"]
        elif previous and previous.field_id == configuration.field_id:
            stages = config["macros"]["goto_field"]["repeat_field_stages"]
        elif configuration.is_rm_field is True:
            stages = config["macros"]["goto_field"]["rm_field_stages"]
        else:
            stages = config["macros"]["goto_field"]["new_field_stages"]

    if stages is not None and len(stages) == 0:
        return command.finish("No stages to run.")

    if with_hartmann and stages is not None and "boss_hartmann" not in stages:
        # Can be added at the end. Macro.reset() will order it.
        stages.append("boss_hartmann")

    macro.reset(
        command,
        stages,
        guider_time=guider_time,
        fixed_rot=fixed_rot,
        fixed_altaz=fixed_altaz,
        fvc_alt=alt,
        fvc_az=az,
        fvc_rot=rot,
        keep_offsets=keep_offsets,
    )
    result = await macro.run()

    if result is False:
        return command.fail()

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-12-25
# @Filename: auto.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
from contextlib import suppress

from hal import config
from hal.exceptions import MacroError
from hal.macros.macro import Macro


class AutoModeMacro(Macro):
    """A macro that runs one iteration of the auto mode."""

    name = "auto"

    __PRECONDITIONS__ = ["prepare"]
    __STAGES__ = ["load", "goto_field", "expose"]
    __CLEANUP__ = ["cleanup"]

    def __init__(self, name: str | None = None):
        super().__init__(name)

        self._preload_design_task: asyncio.Task | None = None

    async def prepare(self):
        """Prepare stage."""

        pass

    def message(self, message: str, level: str = "i"):
        """Emits a message using the ``auto_mode_message`` keyword."""

        self.command.write(level, auto_mode_message=message)

    async def load(self):
        """Load a (pre-)loaded design."""

        jaeger = self.helpers.jaeger

        if jaeger.preloaded:
            self.message("Loading preloaded configuration.")
            await jaeger.from_preloaded(self.command)

        elif jaeger.configuration and jaeger.configuration.observed is False:
            self.message("Found unobserved design.")

        else:
            self.message("No preloaded configurations found. Loading new queue design.")

            if not await jaeger.load_from_queue(self.command):
                raise MacroError("Failed loading design.")

        # Make sure everything all keywords have been emitted and database queries run.
        await asyncio.sleep(1)

        if not jaeger.configuration:
            raise MacroError("Failed loading design. No active configuration.")

        c_id = jaeger.configuration.configuration_id
        d_id = jaeger.configuration.design_id
        self.message(f"Processign configuration {c_id} ({d_id}).")

    async def goto_field(self):
        """Runs the goto-field macro with the appropriate stages."""

        goto = self.helpers.macros["goto_field"]
        expose_macro = self.helpers.macros["expose"]

        if expose_macro.running:
            # Silently skip. The auto expose stage will wait for it.
            return

        if goto.running:
            self.message("goto-field is running. Waiting for it to complete.", "w")
            if not await goto.wait_until_complete():
                raise MacroError("Background goto-field failed. Cancelling auto mode.")
            # Skip goto-field since it's already been done.
            return

        configuration = self.helpers.jaeger.configuration

        if configuration is None:
            return MacroError("No configuration loaded.")
        elif configuration.cloned is True:
            stages = config["macros"]["goto_field"]["cloned_stages"]
        elif configuration.new_field is False:
            stages = config["macros"]["goto_field"]["repeat_field_stages"]
        elif configuration.is_rm_field is True:
            stages = config["macros"]["goto_field"]["rm_field_stages"]
        else:
            stages = config["macros"]["goto_field"]["new_field_stages"]

        if len(stages) == 0:  # For cloned designs.
            self.message("Skipping goto-field.")
            return

        # Check if this configuration has already been marked as "goto_complete",
        # which means the goto-field routine has run at least until just before
        # the acquire stage.
        if configuration.goto_complete:
            self.message("goto-field has already run for this configuration.")
            return

        self.message("Running goto-field.")
        goto.reset(self.command, stages)
        if not await goto.run():
            raise MacroError("Goto field failed during auto mode.")

    async def expose(self):
        """Exposes the camera and schedules next design."""

        expose_macro = self.helpers.macros["expose"]

        if expose_macro.running:
            self.message("expose is running. Waiting for it to complete.", "w")
            if not await expose_macro.wait_until_complete():
                raise MacroError("Background expose failed. Cancelling auto mode.")
            # Skip expose since it's already been done.
            # This will not load a new design but the user can do it while exposing.
            return

        # Make sure guiding is good enough to start exposing.
        if not self.helpers.cherno.is_guiding():
            raise MacroError("Guider is not running. Cannot expose.")

        target_rms = self.config["target_rms"]
        if not self.helpers.cherno.guiding_at_rms(target_rms):
            self.message("RMS not reched yet. Waiting for guider to converge.")
            try:
                await self.helpers.cherno.wait_for_rms(target_rms, max_wait=180)
            except asyncio.TimeoutError:
                raise MacroError("Timed out waiting for guider to converge.")

        # Calculate expose time and schedule preloading a design.
        count = self.config["count"]
        exptime = config["macros"]["expose"]["fallback"]["exptime"]
        flushing = config["durations"]["boss"][self.actor.observatory]["flushing"]
        readout = config["durations"]["boss"][self.actor.observatory]["readout"]
        total_time = (exptime + flushing + readout) * count - readout

        wait_design_load = int(total_time - 180)
        self.message(f"Scheduling next design preload in {wait_design_load} s.", "d")
        await self._preload_design(wait_design_load)

        self.message("Exposing cameras.")
        expose_macro.reset(self.command, count_boss=count)
        if not await expose_macro.run():
            raise MacroError("Expose failed during auto mode.")

    async def cleanup(self):
        """Cleanup stage."""

        if self.cancelled:
            await self._cancel_preload_task()

    async def _preload_design(self, delay: float):
        """Preloads the next design after a delay."""

        async def _preload_executor(delay: float):
            await asyncio.sleep(delay)
            self.message("Preloading design from the queue.")
            await self.helpers.jaeger.load_from_queue(self.command, preload=True)

        await self._cancel_preload_task()
        self._preload_design_task = asyncio.create_task(_preload_executor(delay))

        # At this point we consider this configuration has been observed.
        if self.helpers.jaeger.configuration:
            self.helpers.jaeger.configuration.observed = True

    async def _cancel_preload_task(self):
        """Cancels an existing preload task."""

        if self._preload_design_task and not self._preload_design_task.done():
            self._preload_design_task.cancel()
            with suppress(asyncio.CancelledError):
                await self._preload_design_task

        self._preload_design_task = None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-01-13
# @Filename: expose.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)


from __future__ import annotations

import asyncio
import os
from contextlib import suppress
from dataclasses import dataclass
from time import time

import numpy

from hal import config
from hal.exceptions import MacroError
from hal.macros import Macro


__all__ = ["ExposeMacro"]


@dataclass
class ExposeParameters:
    """Expose macro parameters."""

    boss_exptime: float | None = config["macros"]["expose"]["fallback"]["exptime"]
    apogee_exptime: float | None = None
    count_apogee: int | None = 1
    count_boss: int | None = 1
    pairs: bool = True
    dither: bool = True
    initial_apogee_dither: str = "A"
    readout_matching: bool = True


@dataclass
class BossExposure:
    """Parameters for a BOSS exposure."""

    n: int
    exptime: float
    actual_exptime: float
    read_sync: bool = True
    done: bool = False


@dataclass
class ApogeeExposure:
    """Parameters for an APOGEE exposure."""

    n: int
    exptime: float
    dither_position: str
    done: bool = False


class ExposeHelper:
    """Track exposure status, add/remove exposures, etc."""

    def __init__(self, macro: ExposeMacro, **opts):
        self.macro = macro

        self.observatory = os.environ.get("OBSERVATORY", "APO").upper()

        # Currently running exposure.
        self.n_apogee = 0
        self.n_boss = 0

        # Information about each exposure to take.
        self.apogee_exps: list[ApogeeExposure] = []
        self.boss_exps: list[BossExposure] = []

        self._apogee_exp_start_time: float = 0
        self._boss_exp_start_time: float = 0

        self.interval: float = 10
        self._monitor_task: asyncio.Task | None = None

        self.params: ExposeParameters = ExposeParameters()
        self.update_params(**opts)

        self.refresh()

    def update_params(self, **opts):
        """Update parameters from the macro config."""

        # Exclude options that are not in the dataclass.
        valid_opts = {
            opt: opts[opt]
            for opt in opts
            if opt in ExposeParameters.__dataclass_fields__
        }

        if self.running:
            valid_opts["initial_apogee_dither"] = self.params.initial_apogee_dither

        self.params.__dict__.update(**valid_opts)

        if "expose_boss" not in self.macro.flat_stages:
            self.params.count_boss = None
            self.params.readout_matching = False
        elif "expose_apogee" not in self.macro.flat_stages:
            self.params.count_apogee = None

        return self.params

    async def start(self):
        """Indicate that exposures are starting. Output states on a timer."""

        self._monitor_task = asyncio.create_task(self._monitor())

    async def stop(self):
        """Stops the monitoring task."""

        if self._monitor_task:
            self._monitor_task.cancel()
            with suppress(asyncio.CancelledError):
                await self._monitor_task

        self._monitor_task = None

    async def _monitor(self):
        """Outputs states at intervals."""

        while True:
            self.update_status()
            await asyncio.sleep(self.interval)

    @property
    def running(self):
        """Returns whether the exposures are running."""

        n_apogee = len(self.apogee_exps)
        n_boss = len(self.boss_exps)

        apogee_running = self.n_apogee > 0 and self.n_apogee < n_apogee
        boss_running = self.n_boss > 0 and self.n_boss < n_boss

        return apogee_running or boss_running

    def refresh(self):
        """Refreshes the list of exposures."""

        self._refresh_boss()
        self._refresh_apogee()

        self.update_status()

    def _refresh_boss(self):
        """Refreshes the list of BOSS exposures."""

        count = self.params.count_boss
        exptime = self.params.boss_exptime

        if exptime is None or count is None or count == 0:
            return

        # We don't allow to reduce the number of exposures
        # below the one already being taken.
        if count < self.n_boss:
            count = self.params.count_boss = self.n_boss

        flushing = config["durations"]["boss"][self.observatory]["flushing"]
        readout = config["durations"]["boss"][self.observatory]["readout"]

        if count < len(self.boss_exps):
            # Pop any exposures beyond the count number. Count cannot be
            # smaller than the current exposure.
            while len(self.boss_exps) > count and len(self.boss_exps) >= self.n_boss:
                self.boss_exps.pop()

        # Replace/append exposures, but only for those that have not been executed
        # or are not being executed.
        # Minimum exposure index that we can modify.
        min_exp_idx = 0 if self.n_boss == 0 else self.n_boss
        for ii in range(min_exp_idx, count):
            # Set actual_exptime to zero for now. We'll set it next.
            new = BossExposure(n=ii + 1, exptime=exptime, actual_exptime=0.0)

            # Replace existing exposures, or append if extending.
            if len(self.boss_exps) >= ii + 1:
                self.boss_exps[ii] = new
            else:
                self.boss_exps.append(new)

        # Reset the actual time an exposure will take. This is fine to do
        # at any point (I think). For the last exposure, we do not read the
        # exposure synchronously, and we assume the actual exposure time does
        # not include readout.
        for nexp, exp in enumerate(self.boss_exps):
            exp.actual_exptime = exp.exptime + flushing + readout
            exp.read_sync = True
            if nexp == len(self.boss_exps) - 1:
                exp.actual_exptime -= readout
                exp.read_sync = False

    def _refresh_apogee(self):
        """Refreshes the list of APOGEE exposures."""

        initial_dither = self.params.initial_apogee_dither
        pairs = self.params.pairs
        n_exp = self.params.count_apogee

        if n_exp is None or n_exp == 0:
            return

        # We consider n_exp the number of actual exposures, not dither pairs.
        if pairs:
            n_exp *= 2

        # We don't allow to reduce the number of exposures below the one
        # already being taken. If doing dither pairs, we require completing them.
        if n_exp < self.n_apogee:
            if pairs:
                n_exp = self.n_apogee + self.n_apogee % 2
            else:
                n_exp = self.n_apogee

        exposure_times: list[float] = []

        if self.params.readout_matching:
            # We are matching exposure times. We use the BOSS exposure time as
            # reference to determine the full APOGEE exposure time.

            boss_exptime = self.params.boss_exptime
            flushing: float = config["durations"]["boss"][self.observatory]["flushing"]
            readout: float = config["durations"]["boss"][self.observatory]["readout"]

            if boss_exptime is None or boss_exptime == 0:
                return

            apogee_exptime = boss_exptime + flushing + readout

            # If we are doing pairs, the full exposure time corresponds to a pair.
            if pairs:
                apogee_exptime /= 2.0

            exposure_times = [apogee_exptime] * n_exp

            # The last exposure (or two exposures if doing pairs) must be one BOSS
            # readout shorter. If pairs, we distribute that between the two dither
            # positions.
            if pairs:
                last_exptime = apogee_exptime - readout / 2.0
                exposure_times[-2:] = [last_exptime, last_exptime]
            else:
                exposure_times[-1] = apogee_exptime - readout

        else:
            # We are not matching readout times. Just exposure as many APOGEE
            # exposures/pairs as needed.

            apogee_exptime = self.params.apogee_exptime
            if apogee_exptime is None or apogee_exptime == 0:
                return

            exposure_times = [apogee_exptime] * n_exp

        # Round up exposure times.
        exposure_times = list(numpy.ceil(exposure_times))

        dither_sequence = ""
        if self.params.dither is False:
            # If disable_dithering, just keep using the current dither position.
            dither_sequence = initial_dither * n_exp
        else:
            # If we are dithering, the sequence starts on the current position
            # and changes every two dithers to minimise how much we move the
            # mechanism, e.g., ABBAABBA.
            current_dither_position = initial_dither
            dither_sequence = current_dither_position
            for i in range(1, n_exp):
                if i % 2 != 0:
                    if dither_sequence[-1] == "A":
                        dither_sequence += "B"
                    else:
                        dither_sequence += "A"
                else:
                    dither_sequence += dither_sequence[-1]

        if n_exp < len(self.apogee_exps):
            n_apogee = self.n_apogee
            # Pop any exposures beyond n_exp. n_exp cannot be
            # smaller than the current exposure.
            while len(self.apogee_exps) > n_exp and len(self.apogee_exps) >= n_apogee:
                self.apogee_exps.pop()

        # Replace/append exposures, but only for those that have not been executed
        # or are not being executed.

        # Minimum exposure index that we can modify. Preserve full dither sets.
        if pairs is False:
            min_exp_idx = 0 if self.n_apogee == 0 else self.n_apogee
        else:
            min_exp_idx = 0 if self.n_apogee == 0 else self.n_apogee + self.n_apogee % 2

        for ii in range(min_exp_idx, n_exp):
            new = ApogeeExposure(
                n=ii + 1,
                exptime=exposure_times[ii],
                dither_position=dither_sequence[ii],
            )

            # Replace existing exposures, or append if extending.
            if len(self.apogee_exps) >= ii + 1:
                self.apogee_exps[ii] = new
            else:
                self.apogee_exps.append(new)

    def yield_apogee(self):
        """Returns an iterator of APOGEE exposures."""

        while self.n_apogee < len(self.apogee_exps):
            self.n_apogee += 1
            self._apogee_exp_start_time = time()
            self.update_status("apogee")
            yield self.apogee_exps[self.n_apogee - 1]

        self.update_status("apogee")
        yield None

    def yield_boss(self):
        """Returns an iterator of BOSS exposures."""

        while self.n_boss < len(self.boss_exps):
            self.n_boss += 1
            self._boss_exp_start_time = time()
            self.update_status("boss")
            yield self.boss_exps[self.n_boss - 1]

        self.update_status("boss")
        yield None

    def update_status(self, instrument: str | None = None):
        """Emits the exposure status keywords."""

        # Incomplete state. Just filling out the easy bits.
        state_apogee: dict = {
            "current": self.n_apogee,
            "n": len(self.apogee_exps),
            "pairs": self.params.pairs,
            "dither": "A",
            "etr": 0.0,
            "total_time": 0.0,
            "timestamp": 0,
        }

        state_boss: dict = {
            "current": self.n_boss,
            "n": len(self.boss_exps),
            "etr": 0.0,
            "total_time": 0.0,
            "timestamp": 0,
        }

        if not hasattr(self.macro, "command"):
            return

        if (instrument is None or instrument == "apogee") and len(self.apogee_exps) > 0:
            exps = self.apogee_exps
            this_exp = exps[self.n_apogee - 1] if self.n_apogee > 0 else None

            if self.n_apogee == 0:
                state_apogee["dither"] = exps[0].dither_position
            else:
                state_apogee["dither"] = exps[self.n_apogee - 1].dither_position

            state_apogee["total_time"] = int(round(sum([exp.exptime for exp in exps])))
            state_apogee["timestamp"] = round(time(), 1)

            n_completed = 0 if self.n_apogee == 0 else self.n_apogee - 1
            if this_exp and this_exp.done:
                n_completed += 1

            etr = state_apogee["total_time"]
            etr -= sum([exps[ii].exptime for ii in range(0, n_completed)])
            if self._apogee_exp_start_time > 0 and this_exp and not this_exp.done:
                exp_elapsed = time() - self._apogee_exp_start_time
                etr -= exp_elapsed
                if etr < 0:
                    etr = 0
            state_apogee["etr"] = int(round(etr))

            self.macro.command.debug(exposure_state_apogee=list(state_apogee.values()))

        if (instrument is None or instrument == "boss") and len(self.boss_exps) > 0:
            exps = self.boss_exps
            this_exp = exps[self.n_boss - 1] if self.n_boss > 0 else None

            total_time = sum([exp.actual_exptime for exp in exps])
            state_boss["total_time"] = int(round(total_time))

            state_boss["timestamp"] = round(time(), 1)

            n_completed = 0 if self.n_boss == 0 else self.n_boss - 1
            if this_exp and this_exp.done:
                n_completed += 1

            etr = state_boss["total_time"]
            etr -= sum([exps[ii].actual_exptime for ii in range(0, n_completed)])
            if self._boss_exp_start_time > 0 and this_exp and not this_exp.done:
                exp_elapsed = time() - self._boss_exp_start_time
                etr -= exp_elapsed
                if etr < 0:
                    etr = 0
            state_boss["etr"] = int(round(etr))

            self.macro.command.debug(exposure_state_boss=list(state_boss.values()))


class ExposeMacro(Macro):
    """Takes a science exposure with APOGEE and/or BOSS."""

    name = "expose"

    __PRECONDITIONS__ = ["prepare"]
    __STAGES__ = [("expose_boss", "expose_apogee")]
    __CLEANUP__ = ["cleanup"]

    expose_helper: ExposeHelper
    _pause_event = asyncio.Event()

    def _reset_internal(self, **opts):
        """Reset the exposure status."""

        self.expose_helper = ExposeHelper(self, **opts)

        # An event blocker for when we wait to pause the execution of the macro.
        if not self._pause_event.is_set():
            self._pause_event.set()

    async def prepare(self):
        """Prepare for exposures and run checks."""

        self.command.debug(expose_is_paused=False)

        do_apogee = "expose_apogee" in self.flat_stages
        do_boss = "expose_boss" in self.flat_stages

        # First check if we are exposing and if we are fail before doing anything else.
        if do_apogee and self.helpers.apogee.is_exposing():
            raise MacroError("APOGEE is already exposing.")

        if do_boss and self.helpers.boss.is_exposing():
            raise MacroError("BOSS is already exposing.")

        if do_apogee and self.command.actor.observatory != "LCO":
            if not self.command.actor.helpers.apogee.gang_helper.at_cartridge():
                raise MacroError("The APOGEE gang connector is not at the cart.")

        # Check that IEB FBI are off.
        try:
            cmd = await self.send_command("jaeger", "ieb status")
            fbi_led1, fbi_led2, *_ = cmd.replies.get("fbi_led")
        except (MacroError, KeyError):
            fbi_led1 = fbi_led2 = 0.0
            self.command.warning("Failed getting FBI levels but moving on.")

        if float(fbi_led1) > 0.1 or float(fbi_led2) > 0.1:
            raise MacroError("FBI LEDs are not off.")

        # Check lamps. They must be turned off manually (but maybe add a parameter?)
        if self.command.actor.observatory == "APO":
            lamp_st = [lamp[0] for lamp in self.helpers.lamps.list_status().values()]
            if any(lamp_st):
                raise MacroError("Some lamps are on.")
        else:
            self.command.warning("Skipping lamps check for now.")

        # Concurrent tasks to run.
        tasks = []
        if self.command.actor.observatory == "APO":
            tasks.append(self.helpers.ffs.open(self.command))

        if do_apogee:
            initial_dither = self.config["initial_apogee_dither"]
            if initial_dither:
                tasks.append(
                    self.helpers.apogee.set_dither_position(
                        self.command,
                        initial_dither,
                    )
                )

            tasks.append(
                self.helpers.apogee.shutter(
                    self.command,
                    open=True,
                    shutter="apogee",
                )
            )

            if self.config["with_fpi"]:
                tasks.append(
                    self.helpers.apogee.shutter(
                        self.command,
                        open=True,
                        shutter="fpi",
                    )
                )

        await asyncio.gather(*tasks)

        # Tell the helper that we're about to start exposing.
        await self.expose_helper.start()

    async def expose_boss(self):
        """Exposes BOSS."""

        # Just in case there's a readout pending from some previous error. Although
        # maybe it's better to have to manually force this. Note that this does not
        # abort or clear an exposure at the ICC level, just the internal tracking
        # in HAL.
        self.helpers.boss.clear_readout()

        await self._pause_event.wait()

        for exposure_info in self.expose_helper.yield_boss():
            if exposure_info is None:
                break

            await self.helpers.boss.expose(
                self.command,
                exposure_info.exptime,
                read_async=True,
            )

            # Except for the final exposure, we wait until the readout is complete.
            if exposure_info.read_sync:
                await self.helpers.boss.readout(self.command)

            exposure_info.done = True

            # We wait at the end of the loop so that if a new exposure is
            # added while we are paused, the next iteration of the loop
            # will yield it.
            await self._pause_event.wait()

            # If we have added exposures while paused and the expose, then
            # this exposure readout will have become sync so we give it another
            # try.
            if exposure_info.read_sync and self.helpers.boss.readout_pending:
                await self.helpers.boss.readout(self.command)

    async def expose_apogee(self):
        """Exposes APOGEE."""

        await self._pause_event.wait()

        for exposure_info in self.expose_helper.yield_apogee():
            if exposure_info is None:
                break

            await self.helpers.apogee.expose(
                self.command,
                exposure_info.exptime,
                exp_type="object",
                dither_position=exposure_info.dither_position,
            )

            exposure_info.done = True

            await self._pause_event.wait()

    async def cleanup(self):
        """Cancel any running exposures."""

        await self.expose_helper.stop()

        # Close the APOGEE cold shutter.
        if "expose_apogee" in self.flat_stages:
            self.command.info("Closing APOGEE shutter.")
            await self.helpers.apogee.shutter(
                self.command,
                open=False,
                shutter="apogee",
            )

        if self.helpers.apogee.is_exposing():
            self.command.warning("APOGEE exposure is running. Not cancelling it.")

        if self.helpers.boss.is_exposing():
            if self.helpers.boss.is_reading():
                self.command.info("BOSS is reading.")
            else:
                self.command.warning("BOSS exposure is running. Not cancelling it.")

    async def _pause(self):
        """Pauses the execution of the macro."""

        if self._pause_event.is_set():
            self._pause_event.clear()
            self.command.warning("Pausing execution of the expose macro.")
            self.command.debug(expose_is_paused=True)
        else:
            self.command.warning("Macro is already paused.")
            self.command.debug(expose_is_paused=True)

    async def _resume(self):
        """Resumes the execution of the macro."""

        if not self._pause_event.is_set():
            self._pause_event.set()
            self.command.warning("Resuming execution of the expose macro.")
            self.command.debug(expose_is_paused=False)
        else:
            self.command.warning("Macro is already running.")
            self.command.debug(expose_is_paused=False)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-09-27
# @Filename: scripts.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio
import pathlib
from dataclasses import dataclass

from typing import TYPE_CHECKING

import hal
from hal.actor import HALCommandType


if TYPE_CHECKING:
    from clu import Command

    from hal.actor import HALActor


__all__ = ["Scripts"]


@dataclass
class Scripts:
    actor: HALActor
    path: pathlib.Path

    def __post_init__(self):
        self.path = pathlib.Path(self.path)
        if not self.path.is_absolute():
            self.path = pathlib.Path(hal.__file__).parent / self.path

        self.running: dict[str, asyncio.Task] = {}

    def list_scripts(self):
        """Returns a list of script names."""

        return [ff.name.replace(ff.suffix, "") for ff in self.path.glob("*.inp")]

    def get_steps(self, name: str) -> list[tuple[str, str, float | None]]:
        """Returns the list of steps of the script."""

        if name not in self.list_scripts():
            raise ValueError(f"Cannot find script {name}.")

        lines = (self.path / (name + ".inp")).read_text().splitlines()

        steps = []
        for line in lines:
            if line.strip().startswith("#"):
                continue
            parts = line.strip().split()
            try:
                timeout = float(parts[0])
                actor = parts[1]
                command_string = " ".join(parts[2:])

            except ValueError:
                actor = parts[0]
                command_string = " ".join(parts[1:])
                timeout = None
            steps.append((actor, command_string, timeout))

        return steps

    async def run(self, name: str, command: Command[HALActor] | None = None) -> bool:
        """Runs a script.

        This coroutine creates a task with all the steps to execute and adds it to
        the ``running`` dictionary. The execution can be cancelled by calling `.stop`,
        at which point the task will be cancelled and `.run` will handle the
        cancellation.

        """

        async def run_steps(steps):
            for n, step in enumerate(steps):
                actor, command_string, timeout = step

                if command:
                    step_message = f"{actor} {command_string}"
                    if timeout:
                        step_message = f"{timeout} {step_message}"
                    command.warning(text=f"Script {name}: {step_message}")

                runner = self.actor if command is None else command

                if command:
                    command.debug(
                        script_step=[
                            name,
                            f"{actor} {command_string}",
                            n + 1,
                            len(steps),
                        ]
                    )

                if actor == "sleep":
                    await asyncio.sleep(float(command_string))
                else:
                    await asyncio.wait_for(
                        runner.send_command(actor, command_string),
                        timeout,
                    )

        if name in self.running:
            raise RuntimeError(f"Script {name} is already running.")

        steps = self.get_steps(name)

        task = asyncio.create_task(run_steps(steps))
        self.running[name] = task
        self._emit_running()

        # Now actually await the task, but be sure to handle cancellation.
        # If there is another kind of error (e.g. a timeout), raise it.
        try:
            await task
        except asyncio.CancelledError:
            return False
        except asyncio.TimeoutError:
            if command:
                command.error(error=f"Script {name}: one of the steps timedout out.")
            return False
        except Exception:
            raise
        finally:
            self.running.pop(name)
            self._emit_running()

        return True

    async def cancel(self, name: str):
        """Cancels a running script."""

        if name not in self.running:
            raise RuntimeError(f"Script {name} is not running.")

        task = self.running[name]
        task.cancel()

    def _emit_running(self, command: HALCommandType | None = None):
        """Emits the ``running_scripts`` keyword."""

        running_scripts = list(self.running.keys())
        if command:
            command.info(running_scripts=running_scripts)
        else:
            self.actor.write("i", running_scripts=running_scripts)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-12-20
# @Filename: boss.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import asyncio

from typing import TYPE_CHECKING

from hal import config
from hal.exceptions import HALError

from . import HALHelper


if TYPE_CHECKING:
    from hal.actor import HALActor, HALCommandType


__all__ = ["BOSSHelper"]


class BOSSHelper(HALHelper):
    """Control for BOSS spectrograph."""

    __readout_pending: bool = False
    __readout_task: asyncio.Task | None = None

    name = "boss"

    def __init__(self, actor: HALActor):
        super().__init__(actor)

    @property
    def readout_pending(self):
        """True if an exposure readout is pending."""

        return (
            self.__readout_pending
            and self.__readout_task
            and not self.__readout_task.done()
        )

    def clear_readout(self):
        """Clears any pending readouts."""

        self.__readout_pending = False

    def get_exposure_state(self):
        """Returns the exposure state."""

        if self.actor.observatory == "APO":
            exposure_state = self.actor.models["boss"]["exposureState"]

            if exposure_state is None or None in exposure_state.value:
                raise ValueError("Unknown BOSS exposure state.")

            return exposure_state.value[0].lower()

        else:
            exposure_state = self.actor.models["yao"]["sp2_status_names"]

            if exposure_state is None or None in exposure_state.value:
                raise ValueError("Unknown BOSS exposure state.")

            return exposure_state

    def is_exposing(self):
        """Returns `True` if the BOSS spectrograph is currently exposing."""

        state = self.get_exposure_state()

        if self.actor.observatory == "APO":
            if state in ["idle", "aborted"]:
                return False
        else:
            if "IDLE" in state.value and "READOUT_PENDING" not in state.value:
                return False

        return True

    def is_reading(self):
        """Returns `True` if the camera is reading."""

        state = self.get_exposure_state()

        if self.actor.observatory == "APO":
            if state in ["reading", "prereading"]:
                return True
        else:
            if "READING" in state.value or "READOUT_PENDING" not in state.value:
                return True

        return False

    async def expose(
        self,
        command: HALCommandType,
        exp_time: float = 0.0,
        exp_type: str = "science",
        readout: bool = True,
        read_async: bool = False,
    ):
        """Exposes BOSS. If ``readout=False``, does not read the exposure."""

        if self.readout_pending is not False:
            raise HALError(
                "Cannot expose. The camera is exposing or a readout is pending."
            )

        if self.actor.observatory == "APO":
            await self._expose_boss_icc(
                command,
                exp_time=exp_time,
                exp_type=exp_type,
                readout=readout,
                read_async=read_async,
            )
        else:
            await self._expose_yao(
                command,
                exp_time=exp_time,
                exp_type=exp_type,
                readout=readout,
                read_async=read_async,
            )

    async def _expose_boss_icc(
        self,
        command: HALCommandType,
        exp_time: float = 0.0,
        exp_type: str = "science",
        readout: bool = True,
        read_async: bool = False,
    ):
        """Expose using ``bossICC``."""

        timeout = (
            exp_time
            + config["timeouts"]["expose"]
            + config["timeouts"]["boss_icc_flushing"]
        )

        command_parts = [f"exposure {exp_type}"]

        if exp_type != "bias":
            command_parts.append(f"itime={round(exp_time, 1)}")

        if readout is False or read_async is True:
            command_parts.append("noreadout")
        else:
            timeout += config["timeouts"]["boss_icc_readout"]

        command_string = " ".join(command_parts)

        await self._send_command(command, "boss", command_string, time_limit=timeout)

        self.__readout_pending = True

        if readout is True and read_async is True:
            # We use a _send_command because readout cannot await on itself.
            self.__readout_task = asyncio.create_task(
                self._send_command(
                    command,
                    "boss",
                    "exposure readout",
                    time_limit=25.0 + config["timeouts"]["boss_icc_readout"],
                )
            )
            return

        self.__readout_pending = not readout

    async def _expose_yao(
        self,
        command: HALCommandType,
        exp_time: float = 0.0,
        exp_type: str = "science",
        readout: bool = True,
        read_async: bool = False,
    ):
        """Expose using ``yao``."""

        timeout = exp_time + config["timeouts"]["expose"]

        if exp_type == "science":
            exp_type = "object"

        command_parts = [f"expose --{exp_type}"]

        if exp_type != "bias":
            command_parts.append(f"{round(exp_time, 1)}")

        if readout is False or read_async is True:
            command_parts.append("--no-readout")
        else:
            timeout += config["timeouts"]["boss_yao_readout"]

        command_string = " ".join(command_parts)

        await self._send_command(command, "yao", command_string, time_limit=timeout)

        self.__readout_pending = True

        if readout is True and read_async is True:
            # We use a _send_command because readout cannot await on itself.
            self.__readout_task = asyncio.create_task(
                self._send_command(
                    command,
                    "yao",
                    "read",
                    time_limit=25.0 + config["timeouts"]["boss_yao_readout"],
                )
            )
            return

        self.__readout_pending = not readout

    async def readout(self, command: HALCommandType):
        """Reads a pending readout."""

        if self.readout_pending is False:
            raise HALError("No pending readout.")

        command.debug("Reading pending BOSS exposure.")

        if self.readout_pending and self.__readout_task:
            await self.__readout_task

        else:
            await self._send_command(
                command,
                "boss",
                "exposure readout",
                time_limit=25.0 + config["timeouts"]["boss_readout"],
            )

        self.clear_readout()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-01-21
# @Filename: actor.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import os

from typing import ClassVar, TypeVar

from clu.legacy import LegacyActor

from hal import __version__
from hal.actor.commands import hal_command_parser


__all__ = ["HALActor", "ActorHelpers"]


T = TypeVar("T", bound="HALActor")


class HALActor(LegacyActor):
    """HAL actor."""

    _instance: ClassVar[HALActor | None] = None
    parser = hal_command_parser

    def __init__(self, *args, **kwargs):
        schema = kwargs.pop("schema", None)
        schema = schema or os.path.join(os.path.dirname(__file__), "../etc/schema.json")

        self.observatory = os.environ.get("OBSERVATORY", "APO")

        if self.observatory == "APO":
            kwargs["models"] += ["tcc", "mcp", "boss"]
        elif self.observatory == "LCO":
            kwargs["models"] += ["yao"]

        super().__init__(*args, schema=schema, **kwargs)

        self.version = __version__

        self.helpers = ActorHelpers(self)

        HALActor._instance = self

    @staticmethod
    def get_instance():
        """Returns the current instance.

        Note that this class is not a proper singleton; if called multiple times
        it will re-initialise and the instance will change. This is an easy way
        to get the instance when needed.

        """

        return HALActor._instance


class ActorHelpers:
    """State helpers."""

    def __init__(self, actor: HALActor):
        from hal.helpers import (
            APOGEEHelper,
            BOSSHelper,
            ChernoHelper,
            FFSHelper,
            HALHelper,
            JaegerHelper,
            LampsHelper,
            Scripts,
            TCCHelper,
        )
        from hal.macros import all_macros

        self.actor = actor
        self.observatory = actor.observatory

        self.apogee = APOGEEHelper(actor)
        self.boss = BOSSHelper(actor)
        self.cherno = ChernoHelper(actor)
        self.ffs = FFSHelper(actor)
        self.jaeger = JaegerHelper(actor)
        self.lamps = LampsHelper(actor)
        self.tcc = TCCHelper(actor)

        self.bypasses: set[str] = set(actor.config["bypasses"])
        self._available_bypasses = ["all"]
        self._available_bypasses += [
            helper.name
            for helper in HALHelper.__subclasses__()
            if helper.name is not None
        ]

        self.scripts = Scripts(actor, actor.config["scripts"][self.observatory])

        self.macros = {macro.name: macro for macro in all_macros}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-01-13
# @Filename: expose.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

from typing import TYPE_CHECKING, cast

import click

from hal import config
from hal.macros.macro import StageType, flatten

from . import hal_command_parser, stages


if TYPE_CHECKING:
    from hal.macros.expose import ExposeMacro

    from .. import HALCommandType


__all__ = ["expose"]


@hal_command_parser.command()
@stages("expose", reset=False)
@click.option(
    "--stop",
    is_flag=True,
    help="Cancels an ongoing expose macro. Does not abort the ongoing exposures.",
)
@click.option(
    "--modify",
    "-m",
    is_flag=True,
    help="Modify a running expose macro. The parameters of the previous expose command "
    "are NOT remembered; all flags must be passed again.",
)
@click.option(
    "--pause",
    is_flag=True,
    help="Pauses the execution of the macro. The current exposures will complete.",
)
@click.option(
    "--resume",
    is_flag=True,
    help="Resumes the execution of the macro.",
)
@click.option(
    "--count",
    "-c",
    type=int,
    help="How many exposures to take. If exposing APOGEE and APOGEE exposure time "
    "is not explicitely defined, the last APOGEE exposure will finish as the "
    "BOSS readout begins.",
)
@click.option(
    "--count-apogee",
    type=int,
    help="How many APOGEE exposures to take. Overrides --count.",
)
@click.option(
    "--count-boss",
    type=int,
    help="How many BOSS exposures to take. Overrides --count.",
)
@click.option(
    "--apogee/--no-apogee",
    " /-A",
    default=True,
    help="Expose APOGEE.",
)
@click.option(
    "--boss/--no-boss",
    " /-B",
    default=True,
    help="Expose BOSS.",
)
@click.option(
    "-t",
    "--exposure-time",
    type=float,
    help="Exposure time, in seconds.",
)
@click.option(
    "-b",
    "--boss-exposure-time",
    type=float,
    help="BOSS exposure time in seconds.",
)
@click.option(
    "-a",
    "--apogee-exposure-time",
    type=float,
    help="APOGEE exposure time in seconds. Disables readout matching.",
)
@click.option(
    "-r",
    "--reads",
    type=int,
    help="Number of APOGEE reads. Incompatible with --apogee-exposure-time.",
)
@click.option(
    "-d",
    "--disable-readout-matching",
    is_flag=True,
    help="Does not try to match exposure times so that the last BOSS readout starts "
    "as APOGEE finishes exposing.",
)
@click.option(
    "--pairs/--no-pairs",
    " /-P",
    default=True,
    help="Do dither pairs or single exposures. If --pairs, the exposure time for "
    "APOGEE refers to each dither and --counts refers to dither pairs.",
)
@click.option(
    "--disable-dithering",
    is_flag=True,
    help="If set, the dither position will not change between exposures.",
)
@click.option(
    "--initial-apogee-dither",
    type=str,
    help="Initial APOGEE dither position.",
)
@click.option(
    "--with-fpi/--without-fpi",
    default=True,
    help="Open the FPI shutter.",
)
async def expose(
    command: HALCommandType,
    macro: ExposeMacro,
    stages: list[StageType] | None,
    stop: bool = False,
    modify: bool = False,
    pause: bool = False,
    resume: bool = False,
    count: int = 1,
    count_apogee: int | None = None,
    count_boss: int | None = None,
    apogee: bool = True,
    boss: bool = True,
    pairs: bool = True,
    disable_dithering: bool = False,
    exposure_time: float | None = None,
    boss_exposure_time: float | None = None,
    apogee_exposure_time: float | None = None,
    reads: int | None = None,
    initial_apogee_dither: str | None = None,
    with_fpi: bool = True,
    disable_readout_matching: bool = False,
):
    """Take science exposures."""

    if (stop or modify or pause or resume) and not macro.running:
        return command.fail("No expose macro currently running.")

    # Check incompatible options.
    if exposure_time and (boss_exposure_time or apogee_exposure_time or reads):
        return command.fail(
            "--exposure-time cannot be used with "
            "--apogee-exposure-time, --boss-exposure-time or --reads."
        )

    if count and (count_apogee or count_boss):
        return command.fail(
            "--count cannot be used with --count-apogee or --count-boss."
        )

    if reads is not None and apogee_exposure_time is not None:
        return command.fail("--reads and --apogee-exposure-time are incompatible.")

    if stop:
        macro.cancel()
        return command.finish("Expose macro has been cancelled")

    if pause and resume:
        return command.fail("--pause and --resume are incompatible.")

    # Handle pause/resume
    if pause:
        await macro._pause()
        return command.finish()

    if resume:
        await macro._resume()
        return command.finish()

    # Convert reads to exposure time.
    if reads is not None:
        apogee_exposure_time = reads * config["durations"]["apogee_read"]

    # Disable readout matching if we are providing exposure times or counts
    # for each instrument.
    if (apogee_exposure_time and boss_exposure_time) or (count_apogee and count_boss):
        disable_readout_matching = True

    # If nothing has been defined explicitely, revert to the defaults.
    if not exposure_time:
        if not boss_exposure_time and not apogee_exposure_time:
            exposure_time = config["macros"]["expose"]["fallback"]["exptime"]
        elif apogee_exposure_time and not boss_exposure_time:
            boss_exposure_time = apogee_exposure_time
            disable_readout_matching = True

    if not count and not count_boss and not count_apogee:
        count = config["macros"]["expose"]["fallback"]["count"]

    if exposure_time:
        boss_exposure_time = exposure_time
        apogee_exposure_time = exposure_time

    if count:
        count_apogee = count_boss = count

    selected_stages = cast(list[StageType], stages or flatten(macro.__STAGES__.copy()))

    if boss is False and "expose_boss" in selected_stages:
        selected_stages.remove("expose_boss")

    if apogee is False and "expose_apogee" in selected_stages:
        selected_stages.remove("expose_apogee")

    if "expose_apogee" not in selected_stages or "expose_boss" not in selected_stages:
        disable_readout_matching = True

    initial_apogee_dither = (
        initial_apogee_dither
        or command.actor.helpers.apogee.get_dither_position()
        or "A"
    )

    params = dict(
        count_apogee=count_apogee,
        count_boss=count_boss,
        pairs=pairs,
        dither=not disable_dithering,
        boss_exptime=boss_exposure_time,
        apogee_exptime=apogee_exposure_time,
        readout_matching=not disable_readout_matching,
    )

    # Handle macro modification.
    if modify:
        command.warning("Modifying running expose macro.")
        macro.expose_helper.update_params(**params)
        macro.expose_helper.refresh()

        return command.finish()

    # From this point on this is a new macro, so it should not be already running.
    if macro.running:
        return command.fail("The expose macro is already running.")

    macro.reset(
        command,
        selected_stages,
        initial_apogee_dither=initial_apogee_dither,
        with_fpi=with_fpi,
        **params,
    )

    result = await macro.run()

    if result is False:
        return command.fail()

    return command.finish()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2021-03-24
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from clu import Command
from sdsstools import get_config, get_logger, get_package_version


NAME = "sdss-hal"


config = get_config("hal")
log = get_logger(NAME)


__version__ = get_package_version(path=__file__, package_name=NAME)


from .actor import HALActor



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['hal', 'hal.actor', 'hal.actor.commands', 'hal.helpers', 'hal.macros']

package_data = \
{'': ['*'], 'hal': ['etc/*', 'etc/scripts/apo/*', 'etc/scripts/lco/*']}

install_requires = \
['click-default-group>=1.2.2,<2.0.0',
 'numpy>=1.22.1,<2.0.0',
 'sdss-clu>=1.4.0,<2.0.0',
 'sdssdb>=0.6.0,<0.7.0',
 'sdsstools>=1.0.0,<2.0.0']

entry_points = \
{'console_scripts': ['hal = hal.__main__:hal']}

setup_kwargs = {
    'name': 'sdss-hal',
    'version': '0.6.0b4',
    'description': 'High-level observing tool for SDSS-V (replaces SOP)',
    'long_description': '# HAL\n\n![Versions](https://img.shields.io/badge/python->3.9-blue)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Documentation Status](https://readthedocs.org/projects/sdss-hal/badge/?version=latest)](https://sdss-hal.readthedocs.io/en/latest/?badge=latest)\n[![Test](https://github.com/sdss/hal/actions/workflows/test.yml/badge.svg)](https://github.com/sdss/hal/actions/workflows/test.yml)\n[![codecov](https://codecov.io/gh/sdss/hal/branch/main/graph/badge.svg)](https://codecov.io/gh/sdss/hal)\n\nHigh-level observing actor for SDSS-V (replaces SOP)\n\n## Installation\n\nYou should be able to install `HAL` by doing\n\n```console\npip install sdss-hal\n```\n\nTo build from source, use\n\n```console\ngit clone git@github.com:sdss/hal\ncd hal\npip install .\n```\n',
    'author': 'José Sánchez-Gallego',
    'author_email': 'gallegoj@uw.edu',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/sdss/hal',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# @Author: José Sánchez-Gallego (gallegoj@uw.edu)
# @Date: 2022-12-26
# @Filename: __init__.py
# @License: BSD 3-clause (http://www.opensource.org/licenses/BSD-3-Clause)

from __future__ import annotations

import functools

from typing import TYPE_CHECKING

import click

from clu import Command
from clu.parsers.click import command_parser, coro_helper

from hal.macros.macro import StageType, flatten


if TYPE_CHECKING:
    from hal.actor.actor import HALActor


hal_command_parser = command_parser


def stages(macro_name: str, reset: bool = True):
    """A decorator that adds ``--stages`` and ``--list-stages`` options.

    If ``reset=True`` resets the macro with the list of stages to run and passes
    the command and macro to the callback. If ``reset=False`` it does not reset
    the macro and, in addition to command and macro, also passes the list of
    stages. The latter is useful when additional options need to be passsed to
    the macro at reset.

    """

    def _split_stages(ctx, param, values):
        if values is None:
            return None

        _stages = [c.strip() for c in values.split(",")]

        return _stages

    def decorator(f):
        @click.option(
            "--list-stages",
            is_flag=True,
            help="List the stages for this macro.",
        )
        @click.option(
            "-s",
            "--stages",
            type=str,
            metavar="<stages>",
            callback=_split_stages,
            help="Comma-separated list of stages to execute.",
        )
        @functools.wraps(f)
        async def wrapper(
            command: Command[HALActor],
            *args,
            list_stages: bool = False,
            stages: list[StageType] | None = None,
            **kwargs,
        ):
            if macro_name not in command.actor.helpers.macros:
                raise click.BadArgumentUsage(f"Invalid macro {macro_name}")

            macro = command.actor.helpers.macros[macro_name]

            if list_stages is True:
                macro.list_stages(command=command, only_all=True)
                return command.finish()

            if stages is not None:
                for stage in stages:
                    if stage not in flatten(macro.__STAGES__ + macro.__CLEANUP__):
                        raise click.BadArgumentUsage(f"Invalid stage {stage}")

            if reset:
                macro.reset(command, stages)
                return await coro_helper(f, command, macro, *args, **kwargs)
            else:
                return await coro_helper(f, command, macro, stages, *args, **kwargs)

        return functools.update_wrapper(wrapper, f)

    return decorator


from .auto import *
from .bypass import *
from .calibrations import *
from .expose import *
from .goto import *
from .goto_field import *
from .script import *
from .status import *
